"""A Retrieval Evaluator is a class that evaluates the results of a retrieval system.
It receives a set of queries used to retrieve a document and their respective retrieved documents,
and returns a score or a label for each document."""

from __future__ import annotations

from typing import Any, Callable, Type, get_type_hints

from tenacity import RetryError

from ragelo.evaluators.base_evaluator import BaseEvaluator
from ragelo.llm_providers.base_llm_provider import BaseLLMProvider, get_llm_provider
from ragelo.logger import logger
from ragelo.types.configurations import BaseRetrievalEvaluatorConfig
from ragelo.types.evaluables import Document, Evaluable
from ragelo.types.experiment import Experiment
from ragelo.types.formats import LLMInputPrompt, LLMResponseType
from ragelo.types.query import Query
from ragelo.types.results import RetrievalEvaluatorResult
from ragelo.types.types import RetrievalEvaluatorTypes
from ragelo.utils import call_async_fn


class BaseRetrievalEvaluator(BaseEvaluator):
    """
    A base class for retrieval evaluators.
    """

    config: BaseRetrievalEvaluatorConfig
    evaluable_name: str = "Retrieved document"

    def __init__(self, config: BaseRetrievalEvaluatorConfig, llm_provider: BaseLLMProvider):
        super().__init__(config, llm_provider)

    def evaluate(
        self,
        query: Query | str,
        document: Document | str,
        query_metadata: dict[str, Any] | None = None,
        doc_metadata: dict[str, Any] | None = None,
    ) -> RetrievalEvaluatorResult:
        """Evaluates a single query-document pair. Returns the raw answer and the processed answer.
        Args:
            query (Query | str): The query to evaluate.
                If a string is provided, a Query object will be created with the provided query_metadata.
            document (Document | str): The document to evaluate.
                If a string is provided, a Document object will be created with the provided doc_metadata.
            query_metadata (dict[str, Any] | None): The metadata for the query.
            doc_metadata (dict[str, Any] | None): The metadata for the document.
        """
        query = Query.assemble_query(query, query_metadata)
        document = Document.assemble_document(document, query.qid, doc_metadata)
        result = call_async_fn(self.evaluate_async, (query, document))

        if result.exception or result.raw_answer is None or result.answer is None:
            raise ValueError(
                f"Failed to evaluate qid: {query.qid} did: {document.did}",
                f"Exception: {result.exception}",
            )
        return result

    async def evaluate_async(self, eval_sample: tuple[Query, Evaluable]) -> RetrievalEvaluatorResult:
        """
        Evaluates a single query-document pair asynchronously.
        Args:
            eval_sample (tuple[Query, Evaluable]): The query and document to evaluate.
        """
        query, document = eval_sample
        if not isinstance(document, Document):
            type_name = type(document).__name__
            raise ValueError(f"can't evaluate a {type_name} in a Retrieval Evaluator")
        exc = None
        if document.evaluation is not None and not self.config.force:
            return RetrievalEvaluatorResult(
                did=document.did,
                qid=query.qid,
                raw_answer=document.evaluation.raw_answer,
                answer=document.evaluation.answer,
                exception=document.evaluation.exception,
            )
        llm_input = self._build_message(query, document)
        try:
            llm_response = await self.llm_provider.call_async(
                input=llm_input,
                response_schema=self.config.llm_response_schema,
            )
            llm_response = self._process_answer(llm_response)
        except ValueError as e:
            logger.warning(f"Failed to PARSE answer for qid: {query.qid} did: {document.did}")
            try:
                llm_response = LLMResponseType(raw_answer=llm_response.raw_answer, parsed_answer=None)
            except Exception:
                llm_response = LLMResponseType(raw_answer="", parsed_answer=None)
            exc = str(e)
        except Exception as e:
            logger.warning(f"Failed to FETCH answers for qid: {query.qid} did: {document.did}")
            if isinstance(e, RetryError):
                exc = str(e.last_attempt.exception())
            else:
                exc = str(e)
            try:
                llm_response = LLMResponseType(raw_answer=llm_response.raw_answer, parsed_answer=None)
            except Exception:
                llm_response = LLMResponseType(raw_answer="", parsed_answer=None)
        return RetrievalEvaluatorResult(
            qid=query.qid,
            did=document.did,
            raw_answer=llm_response.raw_answer,
            answer=llm_response.parsed_answer,
            exception=exc,
        )

    def _get_tuples_to_evaluate(self, experiment: Experiment) -> list[tuple[Query, Evaluable]]:
        """
        Creates the list of pairs (query, evaluable) to evaluate
        """
        tuples_to_eval = []
        all_tuples = 0
        missing_evaluations = 0
        for q in experiment:
            for d in q.retrieved_docs.values():
                if d.evaluation is None:
                    missing_evaluations += 1
                tuples_to_eval.append((q, d))
                all_tuples += 1
        if missing_evaluations == 0 and not self.config.force:
            logger.info(
                f"All {all_tuples} documents are already evaluated.\n"
                "If you want to re-evaluate them, use the --force flag"
            )
        return tuples_to_eval

    def _build_message(self, query: Query, document: Document) -> LLMInputPrompt:
        context = {"query": query, "document": document}
        user_message = self.user_prompt.render(**context) if self.user_prompt else None
        system_prompt = self.system_prompt.render(**context) if self.system_prompt else None

        return LLMInputPrompt(
            system_prompt=system_prompt,
            user_message=user_message,
        )

    @classmethod
    def from_config(cls, config: BaseRetrievalEvaluatorConfig, llm_provider: BaseLLMProvider):
        return cls(config, llm_provider)

    @classmethod
    def get_config_class(cls) -> Type[BaseRetrievalEvaluatorConfig]:
        return get_type_hints(cls)["config"]


class RetrievalEvaluatorFactory:
    registry: dict[RetrievalEvaluatorTypes, Type[BaseRetrievalEvaluator]] = {}

    @classmethod
    def register(cls, evaluator_name: RetrievalEvaluatorTypes) -> Callable:
        def inner_wrapper(
            wrapped_class: Type[BaseRetrievalEvaluator],
        ) -> Type[BaseRetrievalEvaluator]:
            if evaluator_name in cls.registry:
                logger.debug(f"Overwriting {evaluator_name} in registry")
            cls.registry[evaluator_name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(
        cls,
        evaluator_name: RetrievalEvaluatorTypes,
        llm_provider: BaseLLMProvider | str,
        config: BaseRetrievalEvaluatorConfig | None = None,
        **kwargs,
    ) -> BaseRetrievalEvaluator:
        if evaluator_name not in cls.registry:
            raise ValueError(
                f"Unknown retrieval evaluator {evaluator_name}\nValid options are {list(cls.registry.keys())}"
            )
        if isinstance(llm_provider, str):
            llm_provider_instance = get_llm_provider(llm_provider, **kwargs)
        else:
            llm_provider_instance = llm_provider
        if config is None:
            class_ = cls.registry[evaluator_name]
            type_config = class_.get_config_class()
            valid_keys = [field for field in type_config.model_fields]
            valid_args = {k: v for k, v in kwargs.items() if k in valid_keys}
            required_fields = [arg for arg, info in type_config.model_fields.items() if info.is_required()]
            for field in required_fields:
                if field not in valid_args:
                    raise ValueError(f"Required argument {field} for evaluator {evaluator_name} not provided")
            config = type_config(**valid_args)
        return cls.registry[evaluator_name].from_config(config, llm_provider_instance)


def get_retrieval_evaluator(
    evaluator_name: RetrievalEvaluatorTypes | str | None = None,
    llm_provider: BaseLLMProvider | str = "openai",
    config: BaseRetrievalEvaluatorConfig | None = None,
    **kwargs,
) -> BaseRetrievalEvaluator:
    if evaluator_name is None:
        # get the name from the config
        if config is None:
            raise ValueError("Either the evaluator_name or a config object must be provided")
        evaluator_name = config.evaluator_name
    if isinstance(evaluator_name, str):
        try:
            evaluator_name = RetrievalEvaluatorTypes(evaluator_name)
        except ValueError:
            raise ValueError(f"Unknown retrieval evaluator {evaluator_name}")
    if evaluator_name is None:
        raise ValueError("The evaluator_name must be provided")
    return RetrievalEvaluatorFactory.create(
        evaluator_name,
        llm_provider=llm_provider,
        config=config,
        **kwargs,
    )



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, field_validator, model_validator
from typing_extensions import Self

from ragelo.logger import logger
from ragelo.types.results import EvaluatorResult


class ChatMessage(BaseModel):
    sender: str
    content: str

    def __str__(self) -> str:
        return f"{self.sender}: {self.content}"


class Evaluable(BaseModel):
    """A base class for objects that can be evaluated. Either a retrieved document or an agent answer.
    Args:
        evaluation Optional[EvaluatorResult]: The result of the evaluation.
        metadata Optional[dict[str, Any]]: Metadata that can be templated in the prompt.
    """

    qid: str
    evaluation: EvaluatorResult | None = None
    metadata: dict[str, Any] | None = None

    @field_validator("qid", mode="before")
    def qid_into_string(cls, v):
        if not isinstance(v, str):
            try:
                v = str(v)
            except ValueError:
                raise ValueError("qid must be a string or convertible to a string")
        return v

    def add_metadata(self, metadata: dict[str, Any] | None):
        if not metadata:
            return
        if self.metadata is None:
            self.metadata = {}
        for k in metadata:
            if k in self.metadata:
                logger.warning(
                    f"Metadata {k} for {self.__class__.__name__}"
                    " is being overwritten!\n"
                    f"Old metadata: {self.metadata[k]}\n"
                    f"New metadata: {metadata[k]}\n"
                )
            self.metadata[k] = metadata[k]


class Document(Evaluable):
    """A document retrieved by an agent in response to a query.
    Args:
        did str: The document ID.
        text str: The text of the document.
        retrieved_by dict[str, float]: If a document was retrieved by multiple agents,
            the score attributed by each agent for this document.
    """

    did: str
    text: str
    retrieved_by: dict[str, float] = {}

    @field_validator("did", mode="before")
    def did_into_string(cls, v):
        if not isinstance(v, str):
            try:
                v = str(v)
            except ValueError:
                raise ValueError("did must be a string or convertible to a string")
        return v

    def add_retrieved_by(self, agent: str, score: float | None = None, force: bool = False, exist_ok: bool = False):
        """Adds the score of an agent that retrieved the document."""
        if agent in self.retrieved_by and not force:
            if not exist_ok:
                logger.info(f"Document with did {self.did} already retrieved by agent {agent}")
            return
        if score is None:
            score = 1.0
        self.retrieved_by[agent] = score

    def __str__(self) -> str:
        if len(self.text) > 100:
            return f"{self.did}: {self.text[:100]}..."
        return f"{self.did}: {self.text}"

    @classmethod
    def assemble_document(
        cls,
        document: Self | str,
        qid: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Self:
        """Assembles a Document object from a string or a Document object."""
        if isinstance(document, str):
            if qid is None:
                raise ValueError("qid must be provided if document is a string")
            did = "<no_did>"
            if metadata:
                valid_id_fields = ["did", "doc_id", "document_id", "id", "_id"]
                valid_id_fields = [f for f in valid_id_fields if f in metadata]
                if valid_id_fields:
                    did = metadata[valid_id_fields[0]]
            document = cls(qid=qid, did=did, text=document)
        document.add_metadata(metadata)
        return document

    @staticmethod
    def assemble_documents(
        documents: list,
        qid: str,
        metadata: list[dict[str, Any]] | list[None] | None = None,
    ) -> dict[str, "Document"]:
        """Assembles a list of Document objects from a list of strings or Document objects."""
        assembled_docs: dict[str, Document] = {}
        if metadata and len(documents) != len(metadata):
            raise ValueError("The number of documents and document metadata do not match")
        if not metadata:
            metadata = [None] * len(documents)

        for idx, (doc, m) in enumerate(zip(documents, metadata)):
            doc_obj = Document.assemble_document(doc, qid, m)
            if doc_obj.did == "<no_did>":
                doc_obj.did = f"doc_{idx}"
            assembled_docs[doc_obj.did] = doc_obj
        return assembled_docs


class AgentAnswer(Evaluable):
    """An answer generated by an agent in response to a query.
    Args:
        agent str: The agent that provided the answer.
        text str: The text of the answer.
        conversation Optional[list[ChatMessage]]: The conversation between the user and the agent.
    """

    agent: str
    text: str | None = None
    conversation: list[ChatMessage] | None = None

    @model_validator(mode="before")
    @classmethod
    def one_of_text_or_conversation(cls, values):
        text = values.get("text")
        conversation = values.get("conversation")
        if text is None and conversation is None:
            raise ValueError("Either text or conversation must be provided")

        if text is not None and conversation is not None:
            raise ValueError("Only one of text or conversation must be provided")
        return values

    @classmethod
    def assemble_answer(
        cls,
        answer: Self | str,
        qid: str,
        agent: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> "AgentAnswer":
        """Assembles an AgentAnswer object from a string or an AgentAnswer object."""
        if isinstance(answer, str):
            if agent is None:
                raise ValueError("agent must be provided if answer is a string")
            answer = cls(agent=agent, qid=qid, text=answer)
        answer.add_metadata(metadata)
        return answer


class PairwiseGame(Evaluable):
    """A game to be played between two agent answers.
    Args:
        agent_a_answer AgentAnswer: The answer provided by the first agent.
        agent_b_answer AgentAnswer: The answer provided by the second agent.
    """

    agent_a_answer: AgentAnswer
    agent_b_answer: AgentAnswer



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Optional

import rich
from jinja2 import Template

from ragelo.llm_providers.base_llm_provider import BaseLLMProvider
from ragelo.types.configurations import BaseEvaluatorConfig
from ragelo.types.evaluables import Evaluable
from ragelo.types.experiment import Experiment
from ragelo.types.formats import LLMResponseType
from ragelo.types.query import Query
from ragelo.types.results import EvaluatorResult
from ragelo.utils import call_async_fn, get_pbar


class BaseEvaluator(ABC):
    """
    An abstract class for all evaluators. An evaluator is responsible for evaluating a query and an evaluable
    """

    config: BaseEvaluatorConfig
    system_prompt: Optional[Template] = None
    user_prompt: Template
    evaluable_name: str = "Evaluable"

    def __init__(self, config: BaseEvaluatorConfig, llm_provider: BaseLLMProvider):
        self.config = config
        self.llm_provider = llm_provider
        if config.system_prompt:
            self.system_prompt = config.system_prompt
        if config.user_prompt:
            self.user_prompt = config.user_prompt

    def evaluate_experiment(self, experiment: Experiment, n_threads: int | None = None):
        """
        Trigger the evaluator for all the supported evaluables in the experiment.
        The evaluation is done in asynchronously with the number of threads defined in config.n_processes parameter.
        This can be overwritten by the n_threads parameter.

        Args:
            experiment(Experiment): The experiment to evaluate.
            n_threads(int): The number of threads to use for the evaluation.
                If None, the number of threads defined in the config will be used.
        """
        n_threads = n_threads or self.config.n_processes
        call_async_fn(self._evaluate_experiment_async, experiment, n_threads)

    @abstractmethod
    async def evaluate_async(self, eval_sample: tuple[Query, Evaluable]) -> EvaluatorResult:
        """Evaluate a single query and evaluable asynchronously."""
        raise NotImplementedError

    async def _evaluate_experiment_async(self, experiment: Experiment, n_threads: int = 1):
        tuples_to_eval = self._get_tuples_to_evaluate(experiment)
        pbar = get_pbar(
            len(tuples_to_eval),
            self.config.rich_print,
            desc=f"Evaluating {self.evaluable_name}s",
        )

        awaitables_ended = False
        pending: set[asyncio.Future] = set()
        aws = map(self.evaluate_async, tuples_to_eval)
        aws = iter(aws)
        failed = 0
        evaluations = 0
        while pending or not awaitables_ended:
            while len(pending) < n_threads and not awaitables_ended:
                try:
                    aw = next(aws)
                except StopIteration:
                    awaitables_ended = True
                else:
                    pending.add(asyncio.ensure_future(aw))
            if not pending:
                break
            done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)
            while done:
                evaluation = await done.pop()
                evaluations += 1
                pbar.update()
                if evaluation.exception:
                    failed += 1
                    continue
                experiment.add_evaluation(
                    evaluation,
                    exist_ok=True,
                    force=self.config.force,
                    should_print=self.config.verbose,
                )
        pbar.close()
        if self.config.verbose:
            self._print_failed_evaluations(evaluations, failed)
        experiment.save()

    @abstractmethod
    def _get_tuples_to_evaluate(self, queries: Experiment) -> Sequence[tuple[Query, Evaluable]]:
        raise NotImplementedError

    def _process_answer(self, llm_response: LLMResponseType) -> LLMResponseType:
        """Processes the raw answer returned by the LLM. Should be implemented by the subclass if needed."""
        return llm_response

    def _print_failed_evaluations(self, total_evaluations: int, failed_evaluations: int):
        if self.config.rich_print:
            rich.print("✅ Done!")
            if failed_evaluations > 0:
                rich.print(f"[bold red]Failed evaluations: {failed_evaluations}[/bold red]")
            rich.print(f"[bold green]Total evaluations: {total_evaluations}[/bold green]")
            return
        print("✅ Done!")
        if failed_evaluations > 0:
            print(f"Failed evaluations: {failed_evaluations}")
        print(f"Total evaluations: {total_evaluations}")



from __future__ import annotations

from ragelo.evaluators.retrieval_evaluators import BaseRetrievalEvaluator, RetrievalEvaluatorFactory
from ragelo.llm_providers.base_llm_provider import BaseLLMProvider
from ragelo.types.configurations import FewShotEvaluatorConfig
from ragelo.types.evaluables import Document
from ragelo.types.formats import LLMInputPrompt
from ragelo.types.query import Query
from ragelo.types.types import RetrievalEvaluatorTypes
from ragelo.utils import string_to_template


@RetrievalEvaluatorFactory.register(RetrievalEvaluatorTypes.FEW_SHOT)
class FewShotEvaluator(BaseRetrievalEvaluator):
    config: FewShotEvaluatorConfig
    user_prompt = string_to_template("""
        Query: {{ query.query }}
        Passage: {{ document.text }}""")
    few_shot_assistant_answer = string_to_template("""{"reasoning": "{{reasoning}}", "relevance": {{relevance}}}""")

    def __init__(self, config: FewShotEvaluatorConfig, llm_provider: BaseLLMProvider):
        super().__init__(config, llm_provider)
        if config.user_prompt:
            self.user_prompt = config.user_prompt
        if config.few_shot_assistant_answer:
            self.few_shot_assistant_answer = config.few_shot_assistant_answer
        if config.system_prompt:
            self.system_prompt = config.system_prompt
        self.few_shots = config.few_shots

    def _build_message(self, query: Query, document: Document) -> LLMInputPrompt:
        few_shot_messages = self.__build_few_shot_examples()
        few_shot_messages.append({"role": "user", "content": self.user_prompt.render(query=query, document=document)})
        if self.system_prompt:
            return LLMInputPrompt(
                system_prompt=self.system_prompt.render(query=query, document=document), messages=few_shot_messages
            )

        return LLMInputPrompt(messages=few_shot_messages)

    def __build_few_shot_examples(self) -> list[dict[str, str]]:
        few_shot_messages: list[dict[str, str]] = []
        for idx, few_shot in enumerate(self.few_shots):
            query = Query(query=few_shot.query, qid=f"query_{idx}")
            document = Document(text=few_shot.passage, did=f"doc_{idx}", qid=f"query_{idx}")
            few_shot_messages.append(
                {"role": "user", "content": self.user_prompt.render(query=query, document=document)}
            )
            reasoning = few_shot.reasoning
            relevance = few_shot.relevance
            assistant_message = self.few_shot_assistant_answer.render(reasoning=reasoning, relevance=relevance)
            few_shot_messages.append({"role": "assistant", "content": assistant_message})
        return few_shot_messages



from __future__ import annotations

from collections.abc import Iterator
from typing import Any, overload

from pydantic import BaseModel, field_validator
from typing_extensions import Self

from ragelo.logger import logger
from ragelo.types.evaluables import AgentAnswer, Document, PairwiseGame
from ragelo.types.results import AnswerEvaluatorResult, RetrievalEvaluatorResult


class Query(BaseModel):
    """A user query that can have retrieved documents and agent answers.
    Args:
        qid str: The query ID.
        query str: The query text.
        metadata Optional[dict[str, Any]]: Metadata that can be templated in the prompt.
        retrieved_docs dict[str, Document]: A dictionary of retrieved documents, where the key is the document ID.
        answers list[AgentAnswer]: The list of agent answers.
        pairwise_games list[PairwiseGame]: The list games to be played between agent's answers.
            Generated by the evaluator.
    """

    qid: str
    query: str
    metadata: dict[str, Any] | None = None
    retrieved_docs: dict[str, Document] = {}
    answers: dict[str, AgentAnswer] = {}
    pairwise_games: list[PairwiseGame] = []

    @field_validator("qid", mode="before")
    def qid_into_string(cls, v):
        if not isinstance(v, str):
            try:
                v = str(v)
            except ValueError:
                raise ValueError("qid must be a string or convertible to a string")
        return v

    def add_metadata(self, metadata: dict[str, Any] | None):
        """Adds metadata to the query that may be templated in the prompt.
        Args:
            metadata dict[str, Any]: The metadata to add to the query.
        """
        if not metadata:
            return
        if self.metadata is None:
            self.metadata = {}
        for k in metadata:
            if k in self.metadata:
                logger.warning(
                    f"Metadata {k} for query {self.qid} is being overwritten!\n"
                    f"Old metadata: {self.metadata[k]}\n"
                    f"New metadata: {metadata[k]}\n"
                )
            self.metadata[k] = metadata[k]

    def add_retrieved_docs(
        self,
        docs: list,
        agent: str | None = None,
        force: bool = False,
    ):
        """
        Add a list of retrieved documents to the query.
        Args:
            docs (list[Document | str] | list[Tuple[Document | str], float]): The list of documents
                (either the objects or the texts) to add to the query.
                Optionally, a score can be provided for each document.
            agent (str): The agent that retrieved the documents.
            force (bool): Whether to overwrite existing documents.
        """
        # First, transform all of the strings into Document objects.
        # If an agent name was provided, but no score information exists,
        # assume the list is ranked and assign a score of 1/(n+1) to each document.
        for idx, doc in enumerate(docs):
            if isinstance(doc, tuple):
                doc, score = doc
                if agent is None:
                    raise ValueError(
                        "If you provide a score for each retrieved document, you should also provide an agent name."
                    )
            elif agent is not None:
                score = 1 / (idx + 1)
            else:
                score = None
            self.add_retrieved_doc(doc, score=score, agent=agent, force=force)

    def add_retrieved_doc(
        self,
        doc: Document,
        score: float | None = None,
        agent: str | None = None,
        force: bool = False,
        exist_ok: bool = False,
    ):
        """Add a retrieved document to the query.
        Args:
            doc Document: The document object retrieved for the query.
            score float: The score of the document when retrieved.
            agent str: The agent that retrieved the document.
        """
        doc = self.retrieved_docs.get(doc.did, doc.model_copy(deep=True))
        if agent is not None:
            doc.add_retrieved_by(agent, score, force, exist_ok)
        if doc.did in self.retrieved_docs and not force:
            if not exist_ok:
                logger.info(f"Query {self.qid} already have a document {doc.did} retrieved.")
            return

        self.retrieved_docs[doc.did] = doc

    def add_agent_answer(self, answer: AgentAnswer, force: bool = False, exist_ok: bool = False):
        """Add an answer generated by an agent to the query.
        Args:
            answer AgentAnswer | str: The answer (either the object or the text) to add to the query.
            agent str: The id or name of the agent that generated the answer.
                If not provided, it will be autogenerated.
        """
        agent = answer.agent
        answer = self.answers.get(agent, answer)
        if answer.agent in self.answers and not force:
            if not exist_ok:
                logger.info(f"Answer from agent {answer.agent} already exists in query {self.qid}")
            return
        self.answers[agent] = answer

    @overload
    def add_evaluation(
        self,
        evaluation: RetrievalEvaluatorResult,
        force: bool = False,
        exist_ok: bool = False,
    ) -> bool: ...

    @overload
    def add_evaluation(
        self,
        evaluation: AnswerEvaluatorResult,
        force: bool = False,
        exist_ok: bool = False,
    ) -> bool: ...

    def add_evaluation(
        self,
        evaluation: RetrievalEvaluatorResult | AnswerEvaluatorResult,
        force: bool = False,
        exist_ok: bool = False,
    ) -> bool:
        if isinstance(evaluation, RetrievalEvaluatorResult):
            did = evaluation.did
            if did not in self.retrieved_docs:
                logger.warning(f"Trying to add evaluation for non-retrieved document {did} in query {self.qid}")
            if self.retrieved_docs[did].evaluation is not None and not force:
                if not exist_ok:
                    logger.warning(f"Document {did} in query {self.qid} already has an evaluation.")
                return False
            if self.retrieved_docs[did].evaluation is not None:
                if not exist_ok:
                    logger.info(f"Document {did} in query {self.qid} already has an evaluation. Overwriting.")
            self.retrieved_docs[did].evaluation = evaluation
            return True
        if evaluation.pairwise:
            agent_a = evaluation.agent_a
            agent_b = evaluation.agent_b
            if agent_a not in self.answers:
                logger.warning(
                    "Trying to add a pairwise evaluation for a comparison between agents "
                    f"{agent_a} and {agent_b}, but {agent_a} does not have an answer for query {self.qid}"
                )
            if agent_b not in self.answers:
                logger.warning(
                    "Trying to add a pairwise evaluation for a comparison between agents "
                    f"{agent_a} and {agent_b}, but {agent_b} does not have an answer for query {self.qid}"
                )
            for game in self.pairwise_games:
                if game.agent_a_answer.agent == agent_a and game.agent_b_answer.agent == agent_b:
                    if game.evaluation is not None:
                        if not force:
                            if not exist_ok:
                                logger.warning(
                                    f"Query {self.qid} already has an evaluation for agents {agent_a} and {agent_b}."
                                )
                            return False
                        if not exist_ok:
                            logger.info(
                                f"Query {self.qid} already has an evaluation for agents {agent_a} and {agent_b}. "
                                "Overwriting."
                            )
                    game.evaluation = evaluation
                    return True
            logger.info(
                f"Trying to add a pairwise evaluation for a comparison between agents {agent_a} and {agent_b},"
                f" but no game was found for query {self.qid}. Creating a new game."
            )
            agent_a = evaluation.agent_a
            agent_b = evaluation.agent_b
            if agent_a not in self.answers:
                logger.warning(
                    f"Trying to add a pairwise evaluation for a comparison between agents {agent_a} and {agent_b},"
                    f" but {agent_a} does not have an answer for query {self.qid}"
                )
                return False
            if agent_b not in self.answers:
                logger.warning(
                    f"Trying to add a pairwise evaluation for a comparison between agents {agent_a} and {agent_b},"
                    f" but {agent_b} does not have an answer for query {self.qid}"
                )
                return False
            self.pairwise_games.append(
                PairwiseGame(
                    qid=self.qid,
                    agent_a_answer=self.answers[agent_a],
                    agent_b_answer=self.answers[agent_b],
                    evaluation=evaluation,
                )
            )
            return True
        if evaluation.agent is None:
            raise ValueError("A pointwise AnswerEvaluatorResult must have an agent assigned to it.")
        agent = evaluation.agent
        if agent not in self.answers:
            logger.warning(
                f"Trying to add evaluation for agent {agent} in query {self.qid}, but {agent} does not have an answer."
            )
        if self.answers[agent].evaluation is not None and exist_ok:
            return False
        if self.answers[agent].evaluation is not None and not force:
            logger.warning(f"Agent {agent} in query {self.qid} already has an evaluation.")
            return False
        if self.answers[agent].evaluation is not None:
            logger.info(f"Agent {agent} in query {self.qid} already has an evaluation. Overwriting.")
        self.answers[agent].evaluation = evaluation
        return True

    def get_qrels(
        self,
        relevance_key: str | None = "relevance",
        relevance_threshold: int = 0,
    ) -> dict[str, int]:
        """Get a qrels-formatted dictionary with the relevance of the retrieved documents.
        Args:
            relevance_key str: The key in the answer object that contains an integer with the relevance of the
                document.
            relevance_threshold int: The minimum relevance value to consider a document as relevant.
                Documents with a relevance lower than this value will be considered as 0.
        """
        qrels = {}
        if len(self.retrieved_docs) == 0:
            logger.warning(f"Query {self.qid} does not have any retrieved documents. Returning empty qrels.")
        docs_without_relevance = 0
        for did, document in self.retrieved_docs.items():
            if document.evaluation is None:
                docs_without_relevance += 1
                continue
            answer = document.evaluation.answer
            if isinstance(answer, BaseModel):
                answer = answer.model_dump()
            if isinstance(answer, int):
                relevance = answer
            if isinstance(answer, float):
                relevance = int(answer)
            elif isinstance(answer, str):
                try:
                    relevance = int(answer)
                except ValueError:
                    logger.warning(
                        f"Document {did} has a relevance key ({relevance_key})"
                        f" that cannot be converted to an int ({answer})."
                        " Skipping."
                    )
                    docs_without_relevance += 1
                    continue
            elif isinstance(answer, dict):
                if relevance_key is None or relevance_key not in answer:
                    logger.warning(
                        f"Document {did} does not have a relevance key ({relevance_key}) in the evaluation. Skipping."
                    )
                    docs_without_relevance += 1
                    continue
                # check if the relevance is a number or a str that can be converted to an int
                if not isinstance(answer[relevance_key], int):
                    try:
                        relevance = int(answer[relevance_key])
                    except ValueError:
                        logger.warning(
                            f"Document {did} has a relevance key ({relevance_key})"
                            f" that cannot be converted to an int ({answer[relevance_key]})."
                            " Skipping."
                        )
                        docs_without_relevance += 1
                        continue
                relevance = int(answer[relevance_key])
            else:
                logger.warning(
                    f"Unsupported relevance type {type(answer)} for document {did} in query {self.qid}. Skipping."
                )
                docs_without_relevance += 1
                continue
            qrels[did] = 0 if relevance < relevance_threshold else relevance
        if docs_without_relevance > 0:
            logger.warning(f"Query {self.qid} has {docs_without_relevance} documents without relevance.")
        if docs_without_relevance == len(self.retrieved_docs):
            logger.error(f"Query {self.qid} has no documents without relevance.")
        return qrels

    def get_runs(self, agents: list[str] | None = None) -> dict[str, dict[str, dict[str, float]]]:
        """Get a trec-style run dictionary with the documents retrieved by the agents provided.
        If no agents are provided, all agents will be included.
        Args:
            agents list[str]: The list of agent IDs to include in the run.
        """
        runs: dict[str, dict[str, dict[str, float]]] = {}
        for did, document in self.retrieved_docs.items():
            for agent, score in document.retrieved_by.items():
                if agents is not None and agent not in agents:
                    continue
                if agent not in runs:
                    runs[agent] = {self.qid: {}}
                runs[agent][self.qid][did] = float(score)

        if len(runs) == 0:
            logger.warning(f"Query {self.qid} does not have any retrieved documents with an agent assigned.")
        return runs

    def retrieved_docs_iter(self) -> Iterator[Document]:
        return iter(self.retrieved_docs.values())

    @classmethod
    def assemble_query(cls, query: Self | str, metadata: dict[str, Any] | None = None) -> Self:
        """Assembles a Query object from a Query object or a query text.
        Args:
            query Query | str: The query object or the query text.
            metadata dict[str, Any]: Metadata to add to the query.
        """
        if isinstance(query, Query):
            # If it's already the right type and has no additional metadata, return as-is
            if type(query) is cls and not metadata:
                return query
            # Otherwise, create a new instance of the correct class with the data
            query_data = query.model_dump()
            if metadata:
                if query_data.get("metadata"):
                    query_data["metadata"].update(metadata)
                else:
                    query_data["metadata"] = metadata
            return cls(**query_data)
        return cls(qid="<no_qid>", query=query, metadata=metadata)



from ragelo.evaluators.answer_evaluators import PairwiseAnswerEvaluator
from ragelo.evaluators.retrieval_evaluators import (
    DomainExpertEvaluator,
    RDNAMEvaluator,
    ReasonerEvaluator,
    RetrievalEvaluatorFactory,
)

__all__ = [
    "PairwiseAnswerEvaluator",
    "DomainExpertEvaluator",
    "RDNAMEvaluator",
    "ReasonerEvaluator",
    "RetrievalEvaluatorFactory",
]



from __future__ import annotations

from typing import Any, Type, get_type_hints

from ragelo.logger import logger
from ragelo.types.configurations.agent_ranker_configs import AgentRankerConfig
from ragelo.types.experiment import Experiment
from ragelo.types.types import AgentRankerTypes


class AgentRanker:
    config: AgentRankerConfig

    def __init__(
        self,
        config: AgentRankerConfig,
    ):
        self.config = config
        self.name = self.config.ranker_name

    def run(self, experiment: Experiment) -> Any:
        """Computes a score for each agent in the experiment"""
        raise NotImplementedError

    @classmethod
    def from_config(cls, config: AgentRankerConfig):
        return cls(config)

    def get_agents_ratings(self) -> dict[str, float]:
        """Returns the score of all players"""
        raise NotImplementedError

    @classmethod
    def get_config_class(cls) -> Type[AgentRankerConfig]:
        return get_type_hints(cls)["config"]

    def _flatten_evaluations(self, experiment: Experiment) -> list[tuple[str, str, str]]:
        evaluations = []
        for query in experiment:
            for game in query.pairwise_games:
                if game.evaluation is not None:
                    if isinstance(game.evaluation.answer, dict):
                        winner = game.evaluation.answer["winner"]
                    else:
                        winner = game.evaluation.answer.winner
                    evaluations.append(
                        (
                            game.agent_a_answer.agent,
                            game.agent_b_answer.agent,
                            winner,
                        )
                    )
        return evaluations


class AgentRankerFactory:
    registry: dict[AgentRankerTypes, Type[AgentRanker]] = {}

    @classmethod
    def register(cls, name: AgentRankerTypes):
        def inner_wrapper(wrapped_class: Type[AgentRanker]):
            if name in cls.registry:
                logger.warning(f"Overwriting {name} in Answer Evaluator registry")
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(
        cls,
        ranker_name: AgentRankerTypes,
        config: AgentRankerConfig | None = None,
        **kwargs,
    ) -> AgentRanker:
        if ranker_name not in cls.registry:
            raise ValueError(f"Unknown Agent Ranker {ranker_name}")
        if config is None:
            class_ = cls.registry[ranker_name]
            type_config = class_.get_config_class()
            valid_keys = [field for field in type_config.model_fields]
            valid_args = {k: v for k, v in kwargs.items() if k in valid_keys}
            config = type_config(**valid_args)
        return cls.registry[ranker_name].from_config(config)


def get_agent_ranker(
    ranker_name: AgentRankerTypes | str | None = None,
    config: AgentRankerConfig | None = None,
    **kwargs,
) -> AgentRanker:
    if ranker_name is None:
        if config is None:
            raise ValueError("Either ranker_name or config should be provided")
        ranker_name = config.ranker_name
    if isinstance(ranker_name, str):
        ranker_name = AgentRankerTypes(ranker_name)
    return AgentRankerFactory.create(ranker_name, config, **kwargs)



from ragelo.evaluators.answer_evaluators.base_answer_evaluator import AnswerEvaluatorFactory, BaseAnswerEvaluator
from ragelo.types.configurations import CustomPromptAnswerEvaluatorConfig
from ragelo.types.types import AnswerEvaluatorTypes


@AnswerEvaluatorFactory.register(AnswerEvaluatorTypes.CUSTOM_PROMPT)
class CustomPromptEvaluator(BaseAnswerEvaluator):
    config: CustomPromptAnswerEvaluatorConfig



import pytest
from pydantic import BaseModel

from ragelo.llm_providers.openai_client import OpenAIProvider
from ragelo.types.formats import LLMInputPrompt, LLMResponseType


class AnswerModel(BaseModel):
    keyA: str
    keyB: str


class TestOpenAIProvider:
    @pytest.mark.parametrize(
        "response_schema,raw_answer,parsed_answer",
        [
            (None, "fake response", "fake response"),
            (
                {"keyA": "Value of key a", "KeyB": "Value of key b"},
                '{"keyA": "valueA", "keyB": "valueB"}',
                {"keyA": "valueA", "keyB": "valueB"},
            ),
            (
                AnswerModel,
                '{"keyA": "valueA", "keyB": "valueB"}',
                AnswerModel(keyA="valueA", keyB="valueB"),
            ),
        ],
    )
    def test_response(
        self,
        response_schema,
        raw_answer,
        parsed_answer,
        openai_client_mock,
        openai_client_config,
        monkeypatch,
    ):
        openai_client = OpenAIProvider(config=openai_client_config)
        monkeypatch.setattr(openai_client, "_OpenAIProvider__openai_client", openai_client_mock)

        # Test with just user prompt
        user_prompt = "hello world"
        result = openai_client(
            LLMInputPrompt(user_message=user_prompt),
            response_schema=response_schema,
        )
        assert isinstance(result, LLMResponseType)
        assert result.raw_answer == raw_answer
        is_structured = isinstance(response_schema, type) and issubclass(response_schema, BaseModel)
        if is_structured:
            assert isinstance(result.parsed_answer, BaseModel)
            assert AnswerModel(**result.parsed_answer.model_dump()) == parsed_answer
        else:
            assert result.parsed_answer == parsed_answer

        # Test with system and user prompts
        system_prompt = "hello system"
        user_prompt_2 = "hello openai"
        result = openai_client(
            LLMInputPrompt(user_message=user_prompt_2, system_prompt=system_prompt),
            response_schema=response_schema,
        )

        # Verify the correct API methods were called
        if is_structured:
            assert openai_client_mock.responses.parse.called
            call_args = openai_client_mock.responses.parse.call_args_list
            assert call_args[0][1]["model"] == "fake model"
            assert call_args[0][1]["input"] == user_prompt
            assert call_args[1][1]["input"] == user_prompt_2
            assert call_args[1][1]["instructions"] == system_prompt
        else:
            assert openai_client_mock.responses.create.called
            call_args = openai_client_mock.responses.create.call_args_list
            assert call_args[0][1]["model"] == "fake model"
            assert call_args[0][1]["input"] == user_prompt
            assert call_args[1][1]["input"] == user_prompt_2
            assert call_args[1][1]["instructions"] == system_prompt



import logging

import typer

from ragelo import Experiment, get_agent_ranker, get_answer_evaluator, get_llm_provider, get_retrieval_evaluator
from ragelo.cli.answer_evaluators_cli import app as answer_evaluator_app
from ragelo.cli.args import get_params_from_function
from ragelo.cli.retrieval_evaluator_cli import app as retrieval_evaluator_app
from ragelo.cli.utils import get_path
from ragelo.types import CLIConfig

typer.main.get_params_from_function = get_params_from_function  # type: ignore


app = typer.Typer()


app.add_typer(retrieval_evaluator_app, name="retrieval-evaluator")
app.add_typer(answer_evaluator_app, name="answer-evaluator")


@app.command()
def run_all(config: CLIConfig = CLIConfig(), **kwargs):
    """Run all the commands."""
    # set ragelo's logger to INFO
    logging.getLogger("ragelo").setLevel(logging.INFO)
    config = CLIConfig(**kwargs)

    # Parse the LLM provider and remove it from the kwargs
    llm_provider = get_llm_provider(config.llm_provider_name, **kwargs)

    # Get the absolute paths for the input and output files, and ensure that they exist.
    queries_csv_file = get_path(config.data_dir, config.queries_csv_file)
    documents_file = get_path(config.data_dir, config.documents_csv_file)
    answers_file = get_path(config.data_dir, config.answers_csv_file)
    output_file = get_path(config.data_dir, config.output_file, check_exists=False) if config.output_file else None

    experiment = Experiment(
        experiment_name=config.experiment_name,
        save_path=output_file,
        queries_csv_path=queries_csv_file,
        documents_csv_path=documents_file,
        answers_csv_path=answers_file,
        verbose=config.verbose,
        clear_evaluations=config.force,
        rich_print=config.rich_print,
        cache_evaluations=config.save_results,
    )

    kwargs = config.model_dump()
    kwargs.pop("llm_response_schema", None)

    retrieval_evaluator = get_retrieval_evaluator("reasoner", llm_provider=llm_provider, **kwargs)
    answers_evaluator = get_answer_evaluator("pairwise", llm_provider=llm_provider, **kwargs)
    ranker = get_agent_ranker("elo", **kwargs)

    retrieval_evaluator.evaluate_experiment(experiment)
    answers_evaluator.evaluate_experiment(experiment)
    ranker.run(experiment=experiment)
    experiment.save(output_file)


if __name__ == "__main__":
    app()



"""Evaluator based on the paper: Paul Thomas, Seth Spielman, Nick Craswell and
Bhaskar Mitra. Large language models can accurately predict searcher preferences.
https://arxiv.org/abs/2309.10621
"""

import numpy as np
from pydantic import BaseModel

from ragelo.evaluators.retrieval_evaluators.base_retrieval_evaluator import (
    BaseRetrievalEvaluator,
    RetrievalEvaluatorFactory,
)
from ragelo.llm_providers.base_llm_provider import BaseLLMProvider
from ragelo.types.answer_formats import (
    RDNAMAnswerEvaluatorFormat,
    RDNAMAnswerNoAspects,
    RDNAMMultipleAnnotatorsAnswer,
    RDNAMMultipleAnnotatorsAnswerNoAspects,
)
from ragelo.types.configurations import RDNAMEvaluatorConfig
from ragelo.types.evaluables import Document
from ragelo.types.formats import LLMInputPrompt, LLMResponseType
from ragelo.types.query import Query
from ragelo.types.types import RetrievalEvaluatorTypes
from ragelo.utils import string_to_template


@RetrievalEvaluatorFactory.register(RetrievalEvaluatorTypes.RDNAM)
class RDNAMEvaluator(BaseRetrievalEvaluator):
    config: RDNAMEvaluatorConfig
    system_prompt = string_to_template("""
        {% if annotator_role %}{{ annotator_role }} {% endif %} Given a query and a document, you must provide a score on an integer scale of 0 to 2 with the following meanings:
        2 = highly relevant, very helpful for this query
        1 = relevant, may be partly helpful but might contain other irrelevant content
        0 = not relevant, should never be shown for this query
        Assume that you are writing a report on the subject of the topic. If you would use any of the information contained in the document in such a report, mark it 1. If the document is primarily about the topic, or contains vital information about the topic, mark it 2. Otherwise, mark it 0.
        """)
    user_prompt = string_to_template("""
        # Query
        A person has typed {{ query.query }} into a search engine.
        {% if query.metadata and (query.metadata.description or query.metadata.narrative) %}
        They were looking for: {{ query.metadata.description }}
        {{ query.metadata.narrative }}
        {% endif %}

        # Result
        Consider the following document.
        ---BEGIN DOCUMENT CONTENT---
        {{ document.text }}
        ---END DOCUMENT CONTENT---

        # Instructions
        Split this problem into steps:
        Consider the underlying intent of the search.
        {%- if use_aspects %}
        Measure how well the content matches a likely intent of the query.
        Measure how trustworthy the web page is.
        {%- endif %}
        Consider the aspects above and relative importance of each, and decide on a final overall score.
        {%- if multiple %}
        We asked five search engine raters to evaluate the relevance of the web page for the query.
        Each rater used their own independent judgement.
        {%- endif %}""")

    def __init__(self, config: RDNAMEvaluatorConfig, llm_provider: BaseLLMProvider):
        """Initializes an evaluator based on RDNAM framework."""
        super().__init__(config, llm_provider)
        self._role = self.config.annotator_role if self.config.annotator_role else ""
        if self.config.use_aspects and self.config.use_multiple_annotators:
            self.config.llm_response_schema = RDNAMMultipleAnnotatorsAnswer
        elif self.config.use_aspects:
            self.config.llm_response_schema = RDNAMAnswerEvaluatorFormat
        elif self.config.use_multiple_annotators:
            self.config.llm_response_schema = RDNAMMultipleAnnotatorsAnswerNoAspects
        else:
            self.config.llm_response_schema = RDNAMAnswerNoAspects

    def _build_message(self, query: Query, document: Document) -> LLMInputPrompt:
        context = {
            "query": query,
            "document": document,
            "annotator_role": self.config.annotator_role,
            "use_aspects": self.config.use_aspects,
            "multiple": self.config.use_multiple_annotators,
        }
        return LLMInputPrompt(
            system_prompt=self.system_prompt.render(**context),
            user_message=self.user_prompt.render(**context),
        )

    def _process_answer(self, llm_response: LLMResponseType) -> LLMResponseType:
        parsed = llm_response.parsed_answer
        assert isinstance(self.config.llm_response_schema, type(BaseModel))
        assert isinstance(parsed, self.config.llm_response_schema)

        if self.config.use_multiple_annotators:
            assert isinstance(parsed, RDNAMMultipleAnnotatorsAnswer)
            overall = float(
                np.mean(
                    [
                        parsed.annotator_1.overall,
                        parsed.annotator_2.overall,
                        parsed.annotator_3.overall,
                        parsed.annotator_4.overall,
                        parsed.annotator_5.overall,
                    ]
                )
            )
            if self.config.use_aspects:
                trustworthiness = [
                    parsed.annotator_1.trustworthiness,
                    parsed.annotator_2.trustworthiness,
                    parsed.annotator_3.trustworthiness,
                    parsed.annotator_4.trustworthiness,
                    parsed.annotator_5.trustworthiness,
                ]
                intent_match = [
                    parsed.annotator_1.intent_match,
                    parsed.annotator_2.intent_match,
                    parsed.annotator_3.intent_match,
                    parsed.annotator_4.intent_match,
                    parsed.annotator_5.intent_match,
                ]
                intent_match = float(np.mean(intent_match))  # type: ignore
                trustworthiness = float(np.mean(trustworthiness))  # type: ignore
            else:
                intent_match = None
                trustworthiness = None
        else:
            assert isinstance(parsed, RDNAMAnswerEvaluatorFormat) or isinstance(parsed, RDNAMAnswerNoAspects)
            overall = parsed.overall
            if self.config.use_aspects:
                assert isinstance(parsed, RDNAMAnswerEvaluatorFormat)
                intent_match = parsed.intent_match  # type: ignore
                trustworthiness = parsed.trustworthiness  # type: ignore
            else:
                intent_match = None
                trustworthiness = None

        response = RDNAMAnswerEvaluatorFormat(
            overall=overall,
            intent_match=intent_match,
            trustworthiness=trustworthiness,
        )

        return LLMResponseType(
            raw_answer=llm_response.raw_answer,
            parsed_answer=response,
        )



import json
from unittest.mock import AsyncMock

from pydantic import BaseModel, Field

from ragelo import get_retrieval_evaluator
from ragelo.evaluators.retrieval_evaluators import (
    BaseRetrievalEvaluator,
    CustomPromptEvaluator,
    DomainExpertEvaluator,
    FewShotEvaluator,
    RDNAMEvaluator,
    ReasonerEvaluator,
)
from ragelo.types import Document, Query
from ragelo.types.answer_formats import RDNAMAnswerEvaluatorFormat, RDNAMAnswerNoAspects
from ragelo.types.formats import LLMInputPrompt, LLMResponseType
from ragelo.types.results import RetrievalEvaluatorResult
from ragelo.utils import string_to_template


class RetrievalEvaluator(BaseRetrievalEvaluator):
    def _build_message(self, query: Query, document: Document) -> LLMInputPrompt:
        return LLMInputPrompt(user_message=f"Query: {query.query}\nDocument: {document.text}")


class TestRetrievalEvaluator:
    def test_evaluate_single_answer(self, llm_provider_mock, experiment, base_retrieval_eval_config):
        base_retrieval_eval_config.llm_response_schema = {"score": "float"}
        evaluator = RetrievalEvaluator.from_config(config=base_retrieval_eval_config, llm_provider=llm_provider_mock)
        query = experiment["0"]
        doc = query.retrieved_docs["0"]
        result = evaluator.evaluate(query, doc)
        assert result.raw_answer == '{"score": 1.0}'
        assert result.answer == {"score": 1.0}

        expected_prompt = f"Query: {query.query}\nDocument: {doc.text}"
        call_args = llm_provider_mock.async_call_mocker.call_args_list
        assert call_args[0][0][0].user_message == expected_prompt

    def test_evaluate_experiment(
        self,
        llm_provider_mock,
        experiment,
        base_retrieval_eval_config,
    ):
        base_retrieval_eval_config.llm_response_schema = {"score": "float"}
        evaluator = RetrievalEvaluator.from_config(config=base_retrieval_eval_config, llm_provider=llm_provider_mock)
        evaluator.evaluate_experiment(experiment)
        doc_ids = ["0", "1", "2", "3"]
        qids = ["0", "0", "1", "1"]
        for query in experiment:
            for doc in query.retrieved_docs_iter():
                assert isinstance(doc.evaluation, RetrievalEvaluatorResult)
                assert isinstance(doc.evaluation.raw_answer, str)
                assert isinstance(doc.evaluation.answer, dict)
                assert doc.did == doc_ids.pop(0)
                assert doc.qid == qids.pop(0)

    def test_evaluate_with_text(self, llm_provider_mock, base_retrieval_eval_config):
        base_retrieval_eval_config.llm_response_schema = {"score": "float"}
        evaluator = RetrievalEvaluator.from_config(config=base_retrieval_eval_config, llm_provider=llm_provider_mock)
        result = evaluator.evaluate(document="This is a document", query="This is a query")
        call_args = llm_provider_mock.async_call_mocker.call_args_list
        assert call_args[0][0][0].user_message == "Query: This is a query\nDocument: This is a document"
        assert isinstance(result.answer, dict)
        assert result.answer["score"] == 1.0

    def test_rich_printing(
        self,
        llm_provider_mock,
        base_eval_config,
        experiment,
        capsys,
    ):
        base_eval_config.rich_print = True
        evaluator = RetrievalEvaluator.from_config(
            config=base_eval_config,
            llm_provider=llm_provider_mock,
        )
        evaluator.evaluate_experiment(experiment)
        captured = capsys.readouterr()
        assert "🔎" in captured.out

    def test_get_by_name(self, llm_provider_mock, expert_retrieval_eval_config):
        domain_expert_evaluator = get_retrieval_evaluator(
            "domain_expert",
            llm_provider_mock,
            expert_in=expert_retrieval_eval_config.expert_in,
        )
        assert isinstance(domain_expert_evaluator, DomainExpertEvaluator)
        reasoner_evaluator = get_retrieval_evaluator(
            "reasoner",
            llm_provider_mock,
        )
        assert isinstance(reasoner_evaluator, ReasonerEvaluator)
        rdna_evaluator = get_retrieval_evaluator("RDNAM", llm_provider_mock)

        assert isinstance(rdna_evaluator, RDNAMEvaluator)
        custom_evaluator = get_retrieval_evaluator(
            "custom_prompt",
            llm_provider_mock,
            user_prompt="Query: {{ query.query }} Retrieved document: {{ document.text }}",
        )
        assert isinstance(custom_evaluator, BaseRetrievalEvaluator)


class TestRDNAMEvaluator:
    def test_process_single_answer(self, llm_provider_mock_rdnam, rdnam_config, experiment):
        evaluator = RDNAMEvaluator.from_config(config=rdnam_config, llm_provider=llm_provider_mock_rdnam)
        query = experiment["0"]
        doc = query.retrieved_docs["0"]
        result = evaluator.evaluate(query, doc)
        assert isinstance(result.answer, RDNAMAnswerEvaluatorFormat)
        assert result.answer.intent_match == 1.0
        assert result.answer.trustworthiness == 0.8
        assert result.answer.overall == 1.2
        assert result.raw_answer is not None
        assert "annotator_1" in result.raw_answer
        assert "annotator_2" in result.raw_answer

        call_args = llm_provider_mock_rdnam.async_call_mocker.call_args_list
        assert call_args[0][0][0].system_prompt.startswith("You are a search quality rater evaluating")


class TestReasonerEvaluator:
    def test_process_single_answer(self, llm_provider_reasoner_mock, base_retrieval_eval_config, experiment):
        evaluator = ReasonerEvaluator.from_config(
            config=base_retrieval_eval_config, llm_provider=llm_provider_reasoner_mock
        )
        query = experiment["0"]
        doc = query.retrieved_docs["0"]
        result = evaluator.evaluate(query, doc)
        system_prompt = evaluator.system_prompt.render(query=query, document=doc)
        user_prompt = evaluator.user_prompt.render(query=query, document=doc)
        call_args = llm_provider_reasoner_mock.async_call_mocker.call_args_list

        assert result.answer.score == 2
        assert call_args[0][0][0].user_message == user_prompt
        assert call_args[0][0][0].system_prompt == system_prompt


class TestCustomPromptEvaluator:
    def test_process_single_answer(
        self,
        llm_provider_mock_retrieval,
        custom_prompt_retrieval_eval_config,
        experiment,
    ):
        query = experiment["0"]
        doc = query.retrieved_docs["0"]

        evaluator = CustomPromptEvaluator.from_config(
            config=custom_prompt_retrieval_eval_config,
            llm_provider=llm_provider_mock_retrieval,
        )
        system_prompt = evaluator.system_prompt.render(query=query, document=doc)
        user_prompt = evaluator.user_prompt.render(query=query, document=doc)

        response = evaluator.evaluate(query, doc)
        call_args = llm_provider_mock_retrieval.async_call_mocker.call_args_list

        assert response.answer.score == 2
        assert call_args[0][0][0].user_message == user_prompt
        assert call_args[0][0][0].system_prompt == system_prompt

    def test_process_with_custom_fields(self, llm_provider_mock_retrieval, custom_prompt_retrieval_eval_config):
        custom_prompt_retrieval_eval_config.user_prompt = string_to_template(
            "query: {{ query.query }} doc: {{ document.text }} q_metadata: {{ query.metadata.q_metadata }} d_metadata: {{ document.metadata.d_metadata }}"
        )
        evaluator = get_retrieval_evaluator(
            "custom_prompt",
            llm_provider_mock_retrieval,
            config=custom_prompt_retrieval_eval_config,
        )
        evaluator.evaluate(
            query="this is a query",
            document="this is a document",
            query_metadata={"q_metadata": "q_1"},
            doc_metadata={"d_metadata": "d_1"},
        )
        assert (
            llm_provider_mock_retrieval.async_call_mocker.call_args_list[0][0][0].user_message
            == "query: this is a query doc: this is a document q_metadata: q_1 d_metadata: d_1"
        )


class TestDomainExpertEvaluator:
    def test_process_single_answer(
        self,
        llm_provider_mock_retrieval,
        expert_retrieval_eval_config,
        experiment,
    ):
        evaluator = DomainExpertEvaluator.from_config(
            config=expert_retrieval_eval_config,
            llm_provider=llm_provider_mock_retrieval,
        )
        query = experiment["0"]
        doc = query.retrieved_docs["0"]
        _ = evaluator.evaluate(query, doc)

        assert llm_provider_mock_retrieval.async_call_mocker.call_count == 1
        prompt = llm_provider_mock_retrieval.async_call_mocker.call_args_list[0][0][0]

        assert prompt.system_prompt.startswith("You are a domain expert in")
        assert expert_retrieval_eval_config.expert_in in prompt.system_prompt
        assert expert_retrieval_eval_config.extra_guidelines[0] in prompt.system_prompt
        assert expert_retrieval_eval_config.domain_short in prompt.system_prompt
        assert expert_retrieval_eval_config.company in prompt.system_prompt

        assert prompt.user_message == f"User query:\n{query.query}\n\nDocument passage:\n{doc.text}"


class TestFewShotEvaluator:
    def test_process_single_answer(
        self,
        llm_provider_mock,
        few_shot_retrieval_eval_config,
        experiment,
    ):
        evaluator = FewShotEvaluator.from_config(
            config=few_shot_retrieval_eval_config,
            llm_provider=llm_provider_mock,
        )
        query = experiment["0"]
        doc = query.retrieved_docs["0"]

        _ = evaluator.evaluate(query, doc)
        call_args = llm_provider_mock.async_call_mocker.call_args_list

        # Check that the call was made with the expected structure
        prompt = call_args[0][0][0]
        system_prompt = prompt.system_prompt
        messages = prompt.messages

        # Check system prompt
        assert system_prompt == "System prompt"
        assert len(messages) == len(few_shot_retrieval_eval_config.few_shots) * 2 + 1

        # Check that user prompt contains few-shot examples and the main query
        assert messages[0]["role"] == messages[2]["role"] == "user"
        assert messages[1]["role"] == messages[3]["role"] == "assistant"
        assert "Few shot query 1" in messages[0]["content"]
        assert "Few shot example 1" in messages[0]["content"]
        assert "Few shot query 2" in messages[2]["content"]
        assert "Few shot example 2" in messages[2]["content"]
        assistant_answer_1 = json.loads(messages[1]["content"])
        assistant_answer_2 = json.loads(messages[3]["content"])
        assert assistant_answer_1["reasoning"] == "This is a good document"
        assert assistant_answer_2["reasoning"] == "This is a bad document"
        assert assistant_answer_1["relevance"] == 2
        assert assistant_answer_2["relevance"] == 0
        assert "What is the capital of Brazil?" in messages[-1]["content"]  # Should contain the actual query


class TestReadmeExamples:
    def test_rdnam_example(self, llm_provider_mock_rdnam):
        def side_effect(*args, **kwargs):
            return LLMResponseType(raw_answer='{"overall": 1.0}', parsed_answer=RDNAMAnswerNoAspects(overall=1))

        llm_provider_mock_rdnam.async_call_mocker = AsyncMock(side_effect=side_effect)
        evaluator = get_retrieval_evaluator("RDNAM", llm_provider=llm_provider_mock_rdnam, write_output=False)
        result = evaluator.evaluate(
            query="What is the capital of France?",
            document="Lyon is the second largest city in France.",
        )
        assert isinstance(result.answer, RDNAMAnswerEvaluatorFormat)
        assert result.answer.overall == 1.0
        assert result.answer == RDNAMAnswerEvaluatorFormat(overall=1.0, intent_match=None, trustworthiness=None)
        assert result.raw_answer == '{"overall": 1.0}'

    def test_custom_prompt_evaluator_example(self, llm_provider_mock):
        answer_dict = {
            "relevance": 0,
            "recency": 0,
            "truthfulness": 0,
            "reasoning": (
                "The document is outdated and incorrect. Rio de Janeiro was the capital of "
                "Brazil until 1960 when it was changed to Brasília."
            ),
        }

        system_prompt = """You are a helpful assistant for evaluating the relevance of a retrieved document to a user query.
        You should pay extra attention to how **recent** a document is. A document older than 5 years is considered outdated.

        The answer should be evaluated according to its recency, truthfulness, and relevance to the user query.
        """

        user_prompt = """
        User query: {{ query.query }}

        Retrieved document: {{ document.text }}

        The document has a date of {{ document.metadata.date }}.
        Today is {{ query.metadata.today_date }}.
        """

        expected_user_prompt = string_to_template(
            """
        User query: What is the capital of Brazil?

        Retrieved document: Rio de Janeiro is the capital of Brazil.

        The document has a date of 04-03-1950.
        Today is 08-04-2024.
        """
        ).render()

        class ResponseSchema(BaseModel):
            relevance: int = Field(
                description="An integer, either 0 or 1. 0 if the document is irrelevant, 1 if it is relevant."
            )
            recency: int = Field(
                description="An integer, either 0 or 1. 0 if the document is outdated, 1 if it is recent."
            )
            truthfulness: int = Field(
                description="An integer, either 0 or 1. 0 if the document is false, 1 if it is true."
            )
            reasoning: str = Field(
                description="A short explanation of why you think the document is relevant or irrelevant."
            )

        mocked_answer = LLMResponseType(
            raw_answer=json.dumps(answer_dict), parsed_answer=ResponseSchema(**answer_dict)
        )
        evaluator = get_retrieval_evaluator(
            "custom_prompt",  # name of the retrieval evaluator
            llm_provider=llm_provider_mock,  # Which LLM provider to use
            system_prompt=system_prompt,  # your custom prompt
            user_prompt=user_prompt,  # your custom prompt
            llm_response_schema=ResponseSchema,  # The response schema for the LLM.
        )

        def side_effect(*args, **kwargs):
            return mocked_answer

        llm_provider_mock.async_call_mocker = AsyncMock(side_effect=side_effect)

        result = evaluator.evaluate(
            query="What is the capital of Brazil?",  # The user query
            document="Rio de Janeiro is the capital of Brazil.",  # The retrieved document
            query_metadata={"today_date": "08-04-2024"},  # Some metadata for the query
            doc_metadata={"date": "04-03-1950"},  # Some metadata for the document
        )
        call_args = llm_provider_mock.async_call_mocker.call_args_list
        assert result.raw_answer == json.dumps(answer_dict)
        assert result.answer == ResponseSchema(**answer_dict)
        assert call_args[0][0][0].system_prompt == string_to_template(system_prompt).render()
        assert call_args[0][0][0].user_message == expected_user_prompt



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, model_validator

from ragelo.logger import logger


class LLMInputPrompt(BaseModel):
    system_prompt: str | None = None
    user_message: str | None = None
    messages: list[dict[str, str]] | None = None

    @model_validator(mode="after")
    def check_at_least_one_is_set(self) -> "LLMInputPrompt":
        if not self.system_prompt and not self.user_message and not self.messages:
            raise ValueError("At least one of system_prompt, user_message, or messages must be set")
        if self.system_prompt and not self.messages and not self.user_message:
            logger.warning(
                "You are providing a system prompt, but no messages. "
                "The system prompt will be set as the user_message."
            )
            self.user_message = self.system_prompt
            self.system_prompt = None
        return self


class LLMResponseType(BaseModel):
    raw_answer: str
    parsed_answer: float | str | dict[str, Any] | BaseModel | None



from __future__ import annotations

import re
from typing import Any, Callable, Optional, Type

from jinja2 import Template
from pydantic import BaseModel, Field, field_validator

from ragelo.types.answer_formats import PairWiseAnswerAnswerFormat
from ragelo.types.configurations.base_configs import BaseEvaluatorConfig
from ragelo.types.evaluables import Document
from ragelo.types.types import AnswerEvaluatorTypes


class BaseAnswerEvaluatorConfig(BaseEvaluatorConfig):
    pairwise: bool = Field(default=False, description="Whether or not to the evaluator is pairwise")
    include_relevance_score: bool = Field(
        default=False,
        description="Whether or not to include the document relevance annotations in the prompt",
    )
    include_relevance_reasoning: bool = Field(
        default=True,
        description="Whether or not to include the RetrievalEvaluator reasoning for the document relevance in the prompt.",
    )
    include_raw_documents: bool = Field(
        default=False,
        description="Whether or not to include the raw documents in the prompt",
    )
    factors: Optional[str] = Field(
        default=(
            "the correctness, helpfulness, completeness, accuracy, depth, and level of detail of their responses"
        ),
        description=("A string containing the factors to be used when evaluating an answer. "),
    )
    document_filter: Optional[Callable[[Document], bool]] = Field(
        default=None,
        description=(
            "A function to filter the documents. "
            "It should take a Document object and return a boolean indicating whether the document should be included in the prompt."
        ),
    )
    document_relevance_threshold: Optional[int] = Field(
        default=None,
        description=(
            "The minimum relevance score for a document to be included in the prompt. "
            "By default, all documents are included."
        ),
    )


class PairwiseEvaluatorConfig(BaseAnswerEvaluatorConfig):
    """Configuration for the pairwise evaluator."""

    evaluator_name: AnswerEvaluatorTypes = AnswerEvaluatorTypes.PAIRWISE
    bidirectional: bool = Field(default=True, description="Whether or not to run each game in both directions")
    n_games_per_query: int = Field(default=100, description="Maximum number of games to generate for each query")
    pairwise: bool = Field(default=True, description="Whether or not to the evaluator is pairwise")
    llm_response_schema: Optional[Type[BaseModel] | dict[str, Any]] = Field(
        default=PairWiseAnswerAnswerFormat,
        description="The response schema for the LLM.",
    )

    @field_validator("user_prompt", mode="after")
    def validate_user_prompt(cls, prompt: Template) -> Template:
        src = getattr(prompt, "_ragelo_source", None)
        placeholders = set(m.group(1) for m in re.finditer(r"{{\s*([a-zA-Z_][\w\.]*)\s*}}", src or ""))
        required = {"query.query", "game.agent_a_answer.text", "game.agent_b_answer.text"}
        missing = sorted(required - placeholders)
        if missing:
            raise ValueError(f"The user prompt is missing placeholders: {', '.join(missing)}")
        return prompt


class CustomPairwiseEvaluatorConfig(PairwiseEvaluatorConfig):
    """Configuration for a custom pairwise evaluator."""

    evaluator_name: AnswerEvaluatorTypes = AnswerEvaluatorTypes.CUSTOM_PAIRWISE
    user_prompt: Template = Field(
        description=(
            "User prompt to use for the evaluator. Should contain at least a {{ query.query }}, a {{ game.agent_a_answer.text }}, and a {{ game.agent_b_answer.text }} placeholder."
        )
    )


class CustomPromptAnswerEvaluatorConfig(BaseAnswerEvaluatorConfig):
    evaluator_name: AnswerEvaluatorTypes = AnswerEvaluatorTypes.CUSTOM_PROMPT
    system_prompt: Optional[Template] = Field(
        default_factory=lambda: Template(
            "You are a helpful assistant tasked with evaluating the correctness of answers."
        ),
        description="The system prompt to use for the evaluator.",
    )
    user_prompt: Template = Field(
        default_factory=lambda: Template(
            "retrieved documents: {% for document in documents %}{{document.text}}\n{% endfor %} query: {{query.query}} answer: {{answer.text}}"
        ),
        description=(
            "The prompt to be used to evaluate the documents. It should contain a {{query.query}} and a {{document.text}} and a {{answer.text}} placeholder"
        ),
    )

    @field_validator("user_prompt", mode="after")
    def validate_user_prompt(cls, prompt: Template) -> Template:
        src = getattr(prompt, "_ragelo_source", None)
        placeholders = set(m.group(1) for m in re.finditer(r"{{\s*([a-zA-Z_][\w\.]*)\s*}}", src or ""))
        required = {"query.query", "answer.text"}
        missing = sorted(required - placeholders)
        if missing:
            raise ValueError(f"The user prompt is missing placeholders: {', '.join(missing)}")
        return prompt


class PairwiseDomainExpertEvaluatorConfig(PairwiseEvaluatorConfig):
    evaluator_name: AnswerEvaluatorTypes = AnswerEvaluatorTypes.DOMAIN_EXPERT
    expert_in: str = Field(description="What the LLM should mimic being an expert in.")
    company: Optional[str] = Field(
        default=None,
        description="Name of the company or organization that the user that "
        "submitted the query works for. that the domain belongs to. "
        "(e.g.: ChemCorp, CS Inc.)",
    )



from __future__ import annotations

import json
from typing import Any, Type

from openai import AsyncAzureOpenAI, AsyncOpenAI
from openai.types.responses import ResponseTextConfigParam
from openai.types.shared_params import ResponseFormatJSONObject
from pydantic import BaseModel
from tenacity import retry, stop_after_attempt, wait_random_exponential

from ragelo.llm_providers.base_llm_provider import BaseLLMProvider, LLMProviderFactory
from ragelo.types.configurations import OpenAIConfiguration
from ragelo.types.formats import LLMInputPrompt, LLMResponseType
from ragelo.types.types import LLMProviderTypes


@LLMProviderFactory.register(LLMProviderTypes.OPENAI)
class OpenAIProvider(BaseLLMProvider):
    """A Wrapper over the OpenAI client."""

    config: OpenAIConfiguration
    api_key_env_var: str = "OPENAI_API_KEY"

    def __init__(self, config: OpenAIConfiguration) -> None:
        super().__init__(config)
        self.__openai_client = self.__get_openai_client(config)
        if self.config.model.startswith("gpt-5") or self.config.model.startswith("o"):
            self.config.temperature = None

    @retry(wait=wait_random_exponential(min=1, max=120), stop=stop_after_attempt(1))
    async def call_async(
        self,
        input: LLMInputPrompt,
        response_schema: Type[BaseModel] | dict[str, Any] | None = None,
    ) -> LLMResponseType:
        """Calls the OpenAI API asynchronously.

        Args:
            input: A LLMInputPrompt object containing the system prompt, user message, or a list of messages.
            response_schema: The schema of the response to expect. If the answer_format is STRUCTURED,
                this should be a Pydantic BaseModel class (not instance). If the answer_format is JSON, this should be
                a dictionary with the desired format the answer should be in. Otherwise, the response will be returned as a string.
        Returns:
            The response from the OpenAI Responses API, formatted according to the answer_format. The LLMResponseType.raw_answer  contains the raw LLM response as a string and the LLMResponseType.parsed_answer contains the parsed response. This can be a number or string (if response_schema is None) a dictionary (if response_schema is a dictionary) or a Pydantic BaseModel (if response_schema is a Pydantic BaseModel)).
        """
        llm_input: str | list[dict[str, str]]
        if input.messages:
            llm_input = input.messages
        elif input.user_message:
            llm_input = input.user_message
        else:
            raise ValueError("No input provided")
        parsed_answer: str | dict[str, Any] | BaseModel | None = None
        if input.system_prompt:
            instructions = input.system_prompt
        else:
            instructions = None
        if isinstance(response_schema, type(BaseModel)):
            answers = await self.__openai_client.responses.parse(
                text_format=response_schema,
                input=llm_input,  # type: ignore
                instructions=instructions,
                model=self.config.model,
                temperature=self.config.temperature,
                max_output_tokens=self.config.max_tokens,
            )
            parsed_answer = answers.output_parsed
            raw_answer = answers.output_text
        elif isinstance(response_schema, dict):
            llm_text = {
                "format": {
                    "type": "json_schema",
                    "schema": response_schema,
                    "strict": True,
                }
            }
            answers = await self.__openai_client.responses.create(
                input=llm_input,  # type: ignore
                instructions=instructions,
                model=self.config.model,
                temperature=self.config.temperature,
                max_output_tokens=self.config.max_tokens,
                text=ResponseTextConfigParam(format=ResponseFormatJSONObject(type="json_object")),
            )
            raw_answer = answers.output_text
            try:
                parsed_answer = json.loads(raw_answer)
            except json.JSONDecodeError as e:
                raise ValueError(f"Failed to parse raw JSON answer {raw_answer} as JSON: {e}") from e
        else:
            answers = await self.__openai_client.responses.create(
                input=llm_input,  # type: ignore
                instructions=instructions,
                model=self.config.model,
                temperature=self.config.temperature,
                max_output_tokens=self.config.max_tokens,
            )
            raw_answer = answers.output_text
            parsed_answer = raw_answer

        return LLMResponseType(
            raw_answer=raw_answer,
            parsed_answer=parsed_answer,
        )

    @staticmethod
    def __get_openai_client(openai_config: OpenAIConfiguration) -> AsyncOpenAI:
        if openai_config.api_type == "azure":
            if openai_config.api_base is None:
                raise ValueError("Azure-OpenAI base url (api_base) not found in configuration.")
            return AsyncAzureOpenAI(
                azure_endpoint=openai_config.api_base,
                api_key=openai_config.api_key,
                api_version=openai_config.api_version,
            )
        elif openai_config.api_type == "open_ai" or openai_config.api_type is None:
            return AsyncOpenAI(
                base_url=openai_config.api_base,
                api_key=openai_config.api_key,
                organization=openai_config.org,
            )
        else:
            raise Exception(f"Unknown OpenAI api type: {openai_config.api_type}")



from ragelo.evaluators.retrieval_evaluators import BaseRetrievalEvaluator, RetrievalEvaluatorFactory
from ragelo.types.configurations import ReasonerEvaluatorConfig
from ragelo.types.types import RetrievalEvaluatorTypes
from ragelo.utils import string_to_template


@RetrievalEvaluatorFactory.register(RetrievalEvaluatorTypes.REASONER)
class ReasonerEvaluator(BaseRetrievalEvaluator):
    """
    A document Evaluator that explicitly asks for the reasoning for why a document is relevant.
    """

    config: ReasonerEvaluatorConfig
    system_prompt = string_to_template("""
        You are an impartial expert document annotator, tasked with evaluating if a document contains relevant information to answer a question submitted by a user. 
        Your goal is to evaluate the relevancy of the documents given a user question, and write a concise reasoning for your decision.
            
        You should write one sentence reasoning wether the document is relevant or not for the user question. A document can be:
            - Not relevant: The document is not on topic.
            - Somewhat relevant: The document is on topic but does not fully answer the user question.
            - Very relevant: The document is on topic and answers the user question.
        """)

    user_prompt = string_to_template("""
        [user question]
        {{ query.query }}

        [document content]
        {{ document.text }}
        """)



from __future__ import annotations

import re
from typing import Any, Optional, Type

from jinja2 import Template
from pydantic import BaseModel, ConfigDict, Field, field_validator

from ragelo.types.types import AnswerEvaluatorTypes
from ragelo.utils import string_to_template


class BaseConfig(BaseModel):
    force: bool = Field(
        default=False,
        description="Force the execution of the commands and overwrite any existing files.",
    )
    rich_print: bool = Field(default=False, description="Use rich to print colorful outputs.")
    verbose: bool = Field(
        default=False,
        description="Whether or not to be verbose and print all intermediate steps.",
    )
    llm_provider_name: str = Field(default="openai", description="The name of the LLM provider to be used.")
    use_progress_bar: bool = Field(
        default=True,
        description="Whether or not to show a progress bar while running the evaluations.",
    )
    n_processes: int = Field(
        default=1,
        description="The number of parallel LLM calls to use for the evaluation.",
    )


class BaseEvaluatorConfig(BaseConfig):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    evaluator_name: Optional[str | AnswerEvaluatorTypes] = Field(
        default=None,
        description="The name of the evaluator to use.",
    )
    llm_response_schema: Optional[Type[BaseModel] | dict[str, Any]] = Field(
        default=None,
        description="The response schema for the LLM. If set, should be a json schema or a Pydantic BaseModel (not an instance). Otherwise, the answer will be returned as a string.",
    )
    system_prompt: Optional[Template] = Field(
        default=None,
        description="The system prompt to use for the evaluator.",
    )

    user_prompt: Optional[Template] = Field(
        default=None,
        description="The user prompt to use for the evaluator. Should contain at least a {{ query.query }} placeholder for the query's text.",
    )

    @field_validator("system_prompt", "user_prompt", mode="before")
    def check_system_prompt_and_user_prompt(cls, v: Optional[str | Template]) -> Optional[Template]:
        if isinstance(v, str):
            return string_to_template(v)
        return v

    @field_validator("user_prompt", mode="after")
    def validate_query_and_document_placeholders(cls, prompt: Template) -> Template:
        src = getattr(prompt, "_ragelo_source", None)
        placeholders = set(m.group(1) for m in re.finditer(r"{{\s*([a-zA-Z_][\w\.]*)\s*}}", src or ""))
        if "query.query" not in placeholders:
            raise ValueError("The user prompt must contain a {{query.query}} placeholder")
        return prompt



from __future__ import annotations

import re
from typing import Any, Optional, Type

from jinja2 import Template
from pydantic import BaseModel, Field, field_validator

from ragelo.types.answer_formats import RetrievalAnswerEvaluatorFormat
from ragelo.types.configurations.base_configs import BaseEvaluatorConfig
from ragelo.types.types import RetrievalEvaluatorTypes
from ragelo.utils import string_to_template


class FewShotExample(BaseModel):
    """A few-shot example used in the few-shot retrieval evaluator.
    Args:
        passage str: The passage of the example.
        query str: The query of the example.
        relevance int: The relevance of the example.
        reasoning str: The reasoning behind the relevance
    """

    passage: str
    query: str
    relevance: int
    reasoning: str


class BaseRetrievalEvaluatorConfig(BaseEvaluatorConfig):
    user_prompt: Optional[Template] = Field(
        default=None,
        description="The user prompt to use for the evaluator. Should contain at least a {{ query.query }} and a {{ document.text }} placeholder for the query and the document text.",
    )
    llm_response_schema: Optional[Type[BaseModel] | dict[str, Any]] = Field(default=RetrievalAnswerEvaluatorFormat)

    @field_validator("user_prompt", mode="after")
    def validate_user_prompt(cls, prompt: Optional[Template]) -> Optional[Template]:
        if prompt is None:
            return prompt
        src = getattr(prompt, "_ragelo_source", None)
        placeholders = set(m.group(1) for m in re.finditer(r"{{\s*([a-zA-Z_][\w\.]*)\s*}}", src or ""))
        if "query.query" not in placeholders:
            raise ValueError("The user prompt must contain a {{ query.query }} placeholder")
        if "document.text" not in placeholders:
            raise ValueError("The user prompt must contain a {{ document.text }} placeholder")
        return prompt


class ReasonerEvaluatorConfig(BaseRetrievalEvaluatorConfig):
    evaluator_name: str | RetrievalEvaluatorTypes = RetrievalEvaluatorTypes.REASONER


class DomainExpertEvaluatorConfig(BaseRetrievalEvaluatorConfig):
    evaluator_name: str | RetrievalEvaluatorTypes = RetrievalEvaluatorTypes.DOMAIN_EXPERT
    expert_in: str = Field(description="What the LLM should mimic being an expert in.")
    domain_short: str | None = Field(
        default=None,
        description="A short or alternative name of the domain. (e.g., Chemistry, CS, etc.)",
    )
    company: str | None = Field(
        default=None,
        description="Name of the company or organization that the user that "
        "submitted the query works for. that the domain belongs to. "
        "(e.g.: ChemCorp, CS Inc.)",
    )
    extra_guidelines: list[str] | None = Field(
        default=None,
        description="A list of extra guidelines to be used when reasoning about the relevancy of the document.",
    )


class CustomPromptEvaluatorConfig(BaseRetrievalEvaluatorConfig):
    evaluator_name: str | RetrievalEvaluatorTypes = RetrievalEvaluatorTypes.CUSTOM_PROMPT
    user_prompt: Template = Field(
        description=(
            "The user prompt to be used to evaluate the documents. "
            "It should contain at least a {{query.query}} and a {{document.text}} placeholder"
        ),
    )


class FewShotEvaluatorConfig(BaseRetrievalEvaluatorConfig):
    evaluator_name: str | RetrievalEvaluatorTypes = RetrievalEvaluatorTypes.FEW_SHOT

    few_shots: list[FewShotExample] = Field(
        default_factory=list,
        description="A list of few-shot examples to be used in the prompt",
    )
    few_shot_assistant_answer: Template | None = Field(
        default=None,
        description="The expected answer format from the LLM for each evaluated document "
        "It should contain at least a {{relevance}} placeholder, and, optionally, a {{reasoning}} placeholder",
    )

    @field_validator("few_shot_assistant_answer", mode="before")
    def string_to_template_few_shot_assistant_answer(cls, prompt: str | None) -> Template | None:
        if prompt is None:
            return prompt
        return string_to_template(prompt)

    @field_validator("few_shot_assistant_answer", mode="after")
    def validate_few_shot_assistant_answer(cls, prompt: Template | None) -> Template | None:
        if prompt is None:
            return prompt
        src = getattr(prompt, "_ragelo_source", None)
        placeholders = set(m.group(1) for m in re.finditer(r"{{\s*([a-zA-Z_][\w\.]*)\s*}}", src or ""))
        if "relevance" not in placeholders:
            raise ValueError("The few_shot_assistant_answer must contain a {{ relevance }} placeholder")
        return prompt


class RDNAMEvaluatorConfig(BaseRetrievalEvaluatorConfig):
    evaluator_name: str | RetrievalEvaluatorTypes = RetrievalEvaluatorTypes.RDNAM
    annotator_role: str | None = Field(
        default=None,
        description="A String defining the type of user the LLM should mimic. "
        "(e.g.: 'You are a search quality rater evaluating the relevance "
        "of web pages')",
    )
    use_aspects: bool = Field(
        default=False,
        description="Should the prompt include aspects to get to the final score? "
        "If true, will prompt the LLM to compute scores for M (intent match) "
        "and T (trustworthy) for the document before computing the final score.",
    )
    use_multiple_annotators: bool = Field(
        default=False,
        description="Should the prompt ask the LLM to mimic multiple annotators?",
    )



from ragelo.evaluators.answer_evaluators.base_answer_evaluator import (
    AnswerEvaluatorFactory,
    BaseAnswerEvaluator,
    get_answer_evaluator,
)
from ragelo.evaluators.answer_evaluators.chat_pairwise_evaluator import ChatPairwiseEvaluator
from ragelo.evaluators.answer_evaluators.custom_pairwise_evaluator import CustomPairwiseEvaluator
from ragelo.evaluators.answer_evaluators.custom_prompt_evaluator import CustomPromptEvaluator
from ragelo.evaluators.answer_evaluators.domain_expert_evaluator import PairwiseDomainExpertEvaluator
from ragelo.evaluators.answer_evaluators.pairwise_evaluator import PairwiseAnswerEvaluator

__all__ = [
    "AnswerEvaluatorFactory",
    "BaseAnswerEvaluator",
    "CustomPromptEvaluator",
    "CustomPairwiseEvaluator",
    "PairwiseDomainExpertEvaluator",
    "PairwiseAnswerEvaluator",
    "get_answer_evaluator",
    "ChatPairwiseEvaluator",
]



from ragelo.evaluators.answer_evaluators.base_answer_evaluator import AnswerEvaluatorFactory, BaseAnswerEvaluator
from ragelo.types.configurations import CustomPairwiseEvaluatorConfig
from ragelo.types.types import AnswerEvaluatorTypes


@AnswerEvaluatorFactory.register(AnswerEvaluatorTypes.CUSTOM_PAIRWISE)
class CustomPairwiseEvaluator(BaseAnswerEvaluator):
    """A custom pairwise evaluator that allows for additional customization."""

    config: CustomPairwiseEvaluatorConfig



from ragelo.types.configurations import (
    AgentRankerConfig,
    BaseAnswerEvaluatorConfig,
    BaseConfig,
    EloAgentRankerConfig,
    LLMProviderConfig,
    OllamaConfiguration,
    OpenAIConfiguration,
    PairwiseEvaluatorConfig,
    ReasonerEvaluatorConfig,
)
from ragelo.types.configurations.cli_configs import CLIConfig
from ragelo.types.evaluables import AgentAnswer, Document, PairwiseGame
from ragelo.types.experiment import Experiment
from ragelo.types.query import Query
from ragelo.types.results import AnswerEvaluatorResult, EvaluatorResult, RetrievalEvaluatorResult
from ragelo.types.types import AgentRankerTypes, AnswerEvaluatorTypes, LLMProviderTypes, RetrievalEvaluatorTypes

__all__ = [
    "AgentAnswer",
    "AgentRankerConfig",
    "AgentRankerTypes",
    "CLIConfig",
    "BaseAnswerEvaluatorConfig",
    "BaseConfig",
    "EloAgentRankerConfig",
    "LLMProviderConfig",
    "OllamaConfiguration",
    "OpenAIConfiguration",
    "PairwiseEvaluatorConfig",
    "PairwiseGame",
    "Query",
    "ReasonerEvaluatorConfig",
    "RetrievalEvaluatorResult",
    "RetrievalEvaluatorTypes",
    "AnswerEvaluatorResult",
    "AnswerEvaluatorTypes",
    "EvaluatorResult",
    "Document",
    "LLMProviderTypes",
    "Experiment",
]



from ragelo.evaluators.retrieval_evaluators.base_retrieval_evaluator import (
    BaseRetrievalEvaluator,
    RetrievalEvaluatorFactory,
    get_retrieval_evaluator,
)
from ragelo.evaluators.retrieval_evaluators.custom_prompt_evaluator import CustomPromptEvaluator
from ragelo.evaluators.retrieval_evaluators.domain_expert_evaluator import DomainExpertEvaluator
from ragelo.evaluators.retrieval_evaluators.few_shot_evaluator import FewShotEvaluator
from ragelo.evaluators.retrieval_evaluators.rdnam_evaluator import RDNAMEvaluator
from ragelo.evaluators.retrieval_evaluators.reasoner_evaluator import ReasonerEvaluator

__all__ = [
    "BaseRetrievalEvaluator",
    "CustomPromptEvaluator",
    "DomainExpertEvaluator",
    "FewShotEvaluator",
    "RDNAMEvaluator",
    "ReasonerEvaluator",
    "RetrievalEvaluatorFactory",
    "get_retrieval_evaluator",
]



import json
import warnings

from ragelo import get_answer_evaluator
from ragelo.evaluators.answer_evaluators import (
    BaseAnswerEvaluator,
    ChatPairwiseEvaluator,
    CustomPairwiseEvaluator,
    CustomPromptEvaluator,
    PairwiseAnswerEvaluator,
    PairwiseDomainExpertEvaluator,
)
from ragelo.types.answer_formats import PairWiseAnswerAnswerFormat
from ragelo.types.evaluables import AgentAnswer
from ragelo.types.formats import LLMInputPrompt
from ragelo.types.query import Query
from ragelo.types.results import AnswerEvaluatorResult


def test_get_by_name(llm_provider_mock):
    pairwise_evaluator = get_answer_evaluator("pairwise", llm_provider_mock)
    assert isinstance(pairwise_evaluator, PairwiseAnswerEvaluator)
    custom_evaluator = get_answer_evaluator(
        "custom_prompt",
        llm_provider_mock,
        system_prompt="system prompt",
        user_prompt="Query: {{ query.query }} Answer agent a: {{ answer.text }}",
    )
    assert isinstance(custom_evaluator, CustomPromptEvaluator)
    custom_pairwise_evaluator = get_answer_evaluator(
        "custom_pairwise",
        llm_provider=llm_provider_mock,
        system_prompt="system prompt",
        user_prompt="Query: {{ query.query }} Answer agent a: {{ game.agent_a_answer.text }} Answer agent b: {{ game.agent_b_answer.text }}",
    )
    assert isinstance(custom_pairwise_evaluator, CustomPairwiseEvaluator)
    domain_expert_evaluator = get_answer_evaluator(
        "domain_expert",
        expert_in="computer science",
        llm_provider=llm_provider_mock,
    )
    assert isinstance(domain_expert_evaluator, PairwiseDomainExpertEvaluator)

    chat_pairwise_evaluator = get_answer_evaluator(
        "chat_pairwise",
        llm_provider=llm_provider_mock,
    )
    assert isinstance(chat_pairwise_evaluator, ChatPairwiseEvaluator)


class TestAnswerEvaluator:
    def test_evaluate_single_answer(self, llm_provider_mock, experiment, base_answer_eval_config, answer_eval_format):
        pointwise_evaluator = BaseAnswerEvaluator.from_config(
            config=base_answer_eval_config,
            llm_provider=llm_provider_mock,
        )
        query = experiment["0"]
        answer = query.answers["agent1"]
        result = pointwise_evaluator.evaluate(query, answer)
        assert isinstance(result, AnswerEvaluatorResult)
        assert isinstance(result.answer, answer_eval_format)
        assert result.raw_answer == '{"quality": 1, "trustworthiness": 0, "originality": 0}'
        assert result.exception is None
        assert result.pairwise is False
        assert result.qid == query.qid
        assert result.agent == answer.agent
        call_args = llm_provider_mock.async_call_mocker.call_args_list
        assert len(call_args) == 1
        expected_user_prompt = f"Query: {query.query}\nAnswer: {answer.text}"
        assert call_args[0][0][0].user_message == expected_user_prompt
        assert call_args[0][0][0].system_prompt == base_answer_eval_config.system_prompt.render()

    def test_evaluate_single_game(self, llm_provider_mock, experiment, pairwise_answer_eval_config):
        evaluator = PairwiseAnswerEvaluator.from_config(
            config=pairwise_answer_eval_config, llm_provider=llm_provider_mock
        )
        query = experiment["0"]
        result = evaluator.evaluate(query, answer_a=query.answers["agent1"], answer_b=query.answers["agent2"])

        assert isinstance(result, AnswerEvaluatorResult)
        assert isinstance(result.raw_answer, str)
        assert isinstance(result.answer, PairWiseAnswerAnswerFormat)
        raw_answer = json.loads(result.raw_answer)
        assert raw_answer.keys() == {"answer_a_analysis", "answer_b_analysis", "comparison_reasoning", "winner"}
        assert result.answer.winner == "A"
        assert result.exception is None
        assert result.pairwise is True
        assert result.qid == query.qid
        assert result.agent_a == query.answers["agent1"].agent
        assert result.agent_b == query.answers["agent2"].agent

    def test_evaluate_experiment(
        self, llm_provider_answer_mock, experiment, base_answer_eval_config, answer_eval_format
    ):
        evaluator = BaseAnswerEvaluator.from_config(
            config=base_answer_eval_config, llm_provider=llm_provider_answer_mock
        )
        evaluator.evaluate_experiment(experiment)
        for query in experiment:
            for answer in query.answers.values():
                assert isinstance(answer.evaluation, AnswerEvaluatorResult)
                assert isinstance(answer.evaluation.answer, answer_eval_format)
                assert isinstance(answer.evaluation.raw_answer, str)
                assert answer.evaluation.exception is None
                assert answer.evaluation.pairwise is False
                assert answer.evaluation.qid == query.qid
                assert answer.evaluation.agent == answer.agent

    def test_evaluate_pairwise_experiment(self, llm_provider_mock, experiment, pairwise_answer_eval_config):
        evaluator = PairwiseAnswerEvaluator.from_config(
            config=pairwise_answer_eval_config, llm_provider=llm_provider_mock
        )
        evaluator.evaluate_experiment(experiment)
        for query in experiment:
            assert len(query.pairwise_games) == 2
            for game in query.pairwise_games:
                assert isinstance(game.evaluation, AnswerEvaluatorResult)
                assert isinstance(game.evaluation.answer, PairWiseAnswerAnswerFormat)
                assert isinstance(game.evaluation.raw_answer, str)
                assert game.evaluation.exception is None
                assert game.evaluation.pairwise is True
                assert game.evaluation.qid == query.qid
                assert game.evaluation.agent_a == game.agent_a_answer.agent
                assert game.evaluation.agent_b == game.agent_b_answer.agent


class TestPairwiseAnswerEvaluator:
    def test_evaluate_single_game(
        self,
        llm_provider_mock,
        experiment_with_conversations_and_reasonings,
        pairwise_answer_eval_config,
    ):
        evaluator = PairwiseAnswerEvaluator.from_config(
            config=pairwise_answer_eval_config,
            llm_provider=llm_provider_mock,
        )
        query = experiment_with_conversations_and_reasonings["0"]
        result = evaluator.evaluate(query, answer_a=query.answers["agent1"], answer_b=query.answers["agent2"])
        assert isinstance(result, AnswerEvaluatorResult)
        assert isinstance(result.answer, PairWiseAnswerAnswerFormat)
        assert isinstance(result.raw_answer, str)
        assert result.answer.winner == "A"

        llm_call_args = llm_provider_mock.async_call_mocker.call_args_list
        assert len(llm_call_args) == 1
        assert isinstance(llm_call_args[0][0][0], LLMInputPrompt)
        # Make sure that no games with the same agent were called
        prompt = llm_call_args[0][0][0]
        agent_a_answer = (
            prompt.user_message.split("[The Start of Assistant A's Answer]")[1]
            .split("[The End of Assistant A's Answer]")[0]
            .strip()
        )
        agent_b_answer = (
            prompt.user_message.split("[The Start of Assistant B's Answer]")[1]
            .split("[The End of Assistant B's Answer]")[0]
            .strip()
        )
        assert agent_a_answer != agent_b_answer
        assert agent_a_answer == query.answers["agent1"].text
        assert agent_b_answer == query.answers["agent2"].text

    def test_evaluation_no_documents(self, llm_provider_mock, empty_experiment, pairwise_answer_eval_config):
        evaluator = PairwiseAnswerEvaluator.from_config(
            config=pairwise_answer_eval_config, llm_provider=llm_provider_mock
        )
        empty_experiment.add_query("empty_query", "0")
        empty_experiment.add_agent_answer("answer_a", "agent1", "0")
        empty_experiment.add_agent_answer("answer_b", "agent2", "0")
        query = empty_experiment["0"]
        _ = evaluator.evaluate(query, answer_a=query.answers["agent1"], answer_b=query.answers["agent2"])


class TestCustomPromptEvaluator:
    def test_evaluate_single_answer(
        self, llm_provider_answer_mock, experiment, custom_answer_eval_config, answer_eval_format
    ):
        evaluator = CustomPromptEvaluator.from_config(
            config=custom_answer_eval_config,
            llm_provider=llm_provider_answer_mock,
        )
        query: Query = experiment["0"]
        answer: AgentAnswer = query.answers["agent1"]
        result = evaluator.evaluate(query, answer)
        assert isinstance(result, AnswerEvaluatorResult)
        assert isinstance(result.answer, answer_eval_format)
        assert isinstance(result.raw_answer, str)
        assert isinstance(result.answer.quality, int)
        assert isinstance(result.answer.trustworthiness, int)
        assert isinstance(result.answer.originality, int)
        llm_call_args = llm_provider_answer_mock.async_call_mocker.call_args_list
        documents = list(query.retrieved_docs.values())
        expected_prompt = evaluator.user_prompt.render(
            query=query,
            answer=answer,
            documents=documents,
        )
        assert llm_call_args[0][0][0].user_message == expected_prompt


class TestChatPairwiseEvaluator:
    def test_evaluate_single_game(
        self,
        llm_provider_mock,
        experiment_with_conversations_and_reasonings,
        chat_pairwise_answer_eval_config,
    ):
        evaluator = ChatPairwiseEvaluator.from_config(
            config=chat_pairwise_answer_eval_config,
            llm_provider=llm_provider_mock,
        )
        query = experiment_with_conversations_and_reasonings["0"]

        result = evaluator.evaluate(query, answer_a=query.answers["agent1"], answer_b=query.answers["agent2"])
        assert isinstance(result, AnswerEvaluatorResult)
        assert isinstance(result.answer, PairWiseAnswerAnswerFormat)
        assert isinstance(result.raw_answer, str)
        assert result.answer.winner == "A"
        llm_call_args = llm_provider_mock.async_call_mocker.call_args_list
        assert len(llm_call_args) == 1
        prompt = llm_call_args[0][0][0]
        agent_a_answer = (
            prompt.user_message.split("[The Start of Conversation with Assistant A]")[1]
            .split("[The End of Conversation with Assistant A]")[0]
            .strip()
        )
        agent_b_answer = (
            prompt.user_message.split("[The Start of Conversation with Assistant B]")[1]
            .split("[The End of Conversation with Assistant B]")[0]
            .strip()
        )
        assert agent_a_answer == "\n".join([str(msg) for msg in query.answers["agent1"].conversation])
        assert agent_b_answer == "\n".join([str(msg) for msg in query.answers["agent2"].conversation])


class TestDomainExpertEvaluator:
    def test_evaluate_single_answer(
        self,
        llm_provider_mock,
        experiment_with_conversations_and_reasonings,
        domain_expert_answer_eval_config,
    ):
        evaluator = PairwiseDomainExpertEvaluator.from_config(
            config=domain_expert_answer_eval_config, llm_provider=llm_provider_mock
        )
        query = experiment_with_conversations_and_reasonings["0"]
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            result = evaluator.evaluate(query, answer_a=query.answers["agent1"], answer_b=query.answers["agent2"])
        assert isinstance(result, AnswerEvaluatorResult)
        prompt = llm_provider_mock.async_call_mocker.call_args_list[0][0][0]
        assert isinstance(prompt, LLMInputPrompt)
        assert "You work for" not in prompt.system_prompt

    def test_evaluate_single_answer_with_company(
        self,
        llm_provider_mock,
        experiment_with_conversations_and_reasonings,
        domain_expert_answer_eval_config,
    ):
        domain_expert_answer_eval_config.company = "Zeta Alpha"
        evaluator = PairwiseDomainExpertEvaluator.from_config(
            config=domain_expert_answer_eval_config, llm_provider=llm_provider_mock
        )
        query = experiment_with_conversations_and_reasonings["0"]
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            result = evaluator.evaluate(query, answer_a=query.answers["agent1"], answer_b=query.answers["agent2"])
        assert isinstance(result, AnswerEvaluatorResult)
        prompt = llm_provider_mock.async_call_mocker.call_args_list[0][0][0]
        assert "You work for" in prompt.system_prompt

    def test_evaluate_single_answer_no_documents(
        self,
        llm_provider_mock,
        experiment,
        domain_expert_answer_eval_config,
    ):
        evaluator = PairwiseDomainExpertEvaluator.from_config(
            config=domain_expert_answer_eval_config, llm_provider=llm_provider_mock
        )
        query = experiment["0"]
        result = evaluator.evaluate(query, answer_a=query.answers["agent1"], answer_b=query.answers["agent2"])
        assert isinstance(result, AnswerEvaluatorResult)
        prompt = llm_provider_mock.async_call_mocker.call_args_list[0][0][0]
        assert "You work for" not in prompt.system_prompt



import logging

from rich import print as rich_print

logger = logging.getLogger(__name__)


class CLILogHandler(logging.Handler):
    """A logging handler for the CLI use case"""

    def __init__(self, use_rich: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_rich = use_rich

    def emit(self, record):
        msg = self.format(record)
        if record.levelno == logging.WARNING:
            if self.use_rich:
                rich_print(f"[yellow]{msg}[/yellow]")
            else:
                print("[WARNING] " + msg)
        elif record.levelno == logging.DEBUG:
            if self.use_rich:
                rich_print(f"[blue bold]{msg}[/blue bold]")
            else:
                print("[DEBUG] " + msg)
        elif record.levelno == logging.ERROR:
            if self.use_rich:
                rich_print(f"[bold red]{msg}[/bold red]")
            else:
                print("[ERROR] " + msg)
        else:
            if self.use_rich:
                rich_print(msg)
            else:
                print(msg)



"""Evaluator with a domain expert persona"""

from ragelo.evaluators.retrieval_evaluators import BaseRetrievalEvaluator, RetrievalEvaluatorFactory
from ragelo.types.configurations import DomainExpertEvaluatorConfig
from ragelo.types.evaluables import Document
from ragelo.types.formats import LLMInputPrompt
from ragelo.types.query import Query
from ragelo.types.types import RetrievalEvaluatorTypes
from ragelo.utils import string_to_template


@RetrievalEvaluatorFactory.register(RetrievalEvaluatorTypes.DOMAIN_EXPERT)
class DomainExpertEvaluator(BaseRetrievalEvaluator):
    config: DomainExpertEvaluatorConfig
    system_prompt = string_to_template("""
        You are a domain expert in {{ expert_in }}.{% if company %} You work for {{ company }}.{% endif %} You are tasked with evaluating the performance of a retrieval system for question answering in this domain. The question answering system will be used by internal users{% if company %} of {{ company }}{% endif %}{% if domain_short %} but it also serves some of your external users like {{ domain_short }}{% endif %}. 
        These users are interested in a retrieval system that provides relevant passages based on their questions.

        ## Evaluation Guidelines
        Please think in steps about the relevance of the retrieved document given the original query. Consider the query, the document title (if available), and the document passage. Reason whether the document is not relevant to the query, somewhat relevant to the query, or highly relevant to the query.
        Use the following guidelines to reason about the relevance of the retrieved document:
        - Not Relevant:
            The document contains information that is unrelated, outdated, or completely irrelevant to the query.
            The document may contain some keywords or phrases from the query, but the context and overall meaning do not align with the query's intent.
            The document may be from a different field or time period, rendering it irrelevant to the current query.
        - Somewhat Relevant:
            The document contains some relevant information but lacks comprehensive details or context.
            The document may discuss a related topic or concept but not directly address the query.
            The information in the document is tangentially related to the query, but the primary focus remains different.
        - Highly Relevant:
            The document directly addresses the main points of the query and provides comprehensive and accurate information.
            The document may cite relevant information directly applicable to the query.
            The document may be recent and from the same field as the query, enhancing its relevance.
        General Guidelines:
            - Context Matters: Annotators should evaluate the relevance of documents within the specific context provided by the query. Understanding the 
            - nuances and domain-specific terminology is essential.
            - Content Overlap: Consider the extent of content overlap between the document and the query. Assess whether the document covers the core aspects of the query or only peripheral topics.
            - Neutrality: Base judgments solely on the content's relevance and avoid any personal opinions or biases.
            - Uncertainty: If uncertain about a relevance judgement, annotators default to a lower relevance.
        {% if extra_guidelines %}
        {%- for g in extra_guidelines %}
            - {{ g }}
        {% endfor %}
        {% endif %}
        Given the analysis above, assign a relevance score of 0, 1, or 2 to the retrieved document for this query, where:
        - 0: the document is not relevant to the query
        - 1: the document is somewhat relevant to the query
        - 2: the document is highly relevant to the query

        Respond STRICTLY as a JSON object with the following keys:
        - "reasoning": a concise explanation of your judgment
        - "score": an integer (0, 1, or 2)""")

    user_prompt = string_to_template("""
        User query:
        {{ query.query }}

        Document passage:
        {{ document.text }}""")

    def _build_message(self, query: Query, document: Document) -> LLMInputPrompt:
        context = {
            "query": query,
            "document": document,
            "extra_guidelines": self.config.extra_guidelines or [],
            "expert_in": self.config.expert_in,
            "company": self.config.company,
            "domain_short": self.config.domain_short,
        }
        system_prompt = self.system_prompt.render(**context)
        user_message = self.user_prompt.render(**context)
        return LLMInputPrompt(system_prompt=system_prompt, user_message=user_message)



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, SerializeAsAny, ValidationError, model_validator


class EvaluatorResult(BaseModel):
    """Generic class with the results of an evaluator.
    Args:
        qid str: The query ID to which the result corresponds.
        agent str | None: The agent that provided the answer or retrieved the document.
        raw_answer str | None: The raw answer provided by the LLMProvider used in the evaluator.
        answer Optional[Union[int, str, dict[str, Any]]]: The processed answer provided by the evaluator.
            If the evaluator uses "multi_field_json" as answer_format, this will be a dictionary with evaluation keys.
    """

    qid: str
    agent: str | None = None
    raw_answer: str | None = None
    answer: float | str | dict[str, Any] | SerializeAsAny[BaseModel] | None = None
    exception: str | None = None

    @model_validator(mode="before")
    @classmethod
    def check_agents(cls, v):
        exception = v.get("exception")
        raw_answer = v.get("raw_answer")
        answer = v.get("answer")
        if (raw_answer is None or answer is None) and exception is None:
            raise ValidationError(
                "Either answer or raw_answer must be provided. Otherwise, an exception must be provided."
            )  # type: ignore
        return v


class AnswerEvaluatorResult(EvaluatorResult):
    """The results of an answer evaluator.
    Args:
        agent str | None: The agent that provided the answer. Only used if pairwise=False.
        agent_a str | None: The first agent that provided the answer. Only used if the evaluator is pairwise.
        agent_b str | None: The second agent that provided the answer. Only used if the evaluator is pairwise.
        pairwise bool: Whether the evaluation is pairwise or not.
    """

    agent: str | None = None
    agent_a: str | None = None
    agent_b: str | None = None
    pairwise: bool = False

    @model_validator(mode="before")
    @classmethod
    def check_agents(cls, v):
        agent = v.get("agent")
        agent_a = v.get("agent_a")
        agent_b = v.get("agent_b")
        if agent is None and agent_a is None and agent_b is None:
            raise ValidationError("Either agent or agent_a and agent_b must be provided")  # type: ignore
        if agent_a is not None and agent_b is not None:
            v["pairwise"] = True
        return v


class RetrievalEvaluatorResult(EvaluatorResult):
    """The results of a retrieval evaluator.
    Args:
        did str: The document ID to which the result corresponds.
    """

    did: str


class EloTournamentResult(BaseModel):
    """A class to store the results of an Elo tournament between multiple agents."""

    agents: list[str]
    scores: dict[str, float]
    games_played: dict[str, int]
    wins: dict[str, int]
    loses: dict[str, int]
    ties: dict[str, int]
    std_dev: dict[str, float]
    total_games: int
    total_tournaments: int
